/* $Id: gnugotmail.c,v 1.205 2004/03/09 06:09:11 chrisf Exp $ */

/*
GNU Got Mail - An SMTP/POP3 client for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* Main code for gnugotmail */

#include <PalmOS.h>
#include <System/NetMgr.h>
#include <System/SysEvtMgr.h>
#include <Handspring/HsExt.h>
#include <PhoneLib/PhoneGlobals.h>
#include "callback.h"
#include "SysZLib.h"
#include "gnugotmailRsc.h"
#include "gnugotmail.h"
#include "gnugotmailDB.h"
#include "common.h"
#include "sockets.h"
#include "pop.h"
#include "smtp.h"
#include "timer.h"

/* globals */

Err errno, h_errno;
Int32 AppNetTimeout;
UInt16 AppNetRefnum;
Boolean isnetlibopen=false;

NetSocketRef popsock=0, smtpsock=0;
extern struct spopdata popd;
extern struct ssmtpdata smtpd;
extern UInt32 leavefree;
extern char lastchars[];

Int32 OutboxCount;
char tnum[10], nullchar='\0';
Char AppErrStr[AppErrStrLen];
Boolean PrefsRdirty, PrefsWeredirty;
UInt16 catInbox, catOutbox, catFiled;
DateTimeType dtnow;
UInt32 dtnowseconds;
UInt16 tps;
UInt16 AppsPopupCount;
Handle AppsPopupH;
UInt32 Prefslaunch[6];
Int16 buttonsshown;
Boolean isbackground=false;
Boolean ismymaildb=true;
Int16 backgroundretval=RV_SUCCESS;

#define BSIZ 4096
#define ZBSIZ (BSIZ/4)
unsigned char ibuf[BSIZ], obuf[ZBSIZ];
z_stream pgpz;

/*
 * Heap allocations
 */
MemHandle PrefsRecHandle, PrefsRHandle;
UInt16 PrefsRecIndex;
struct sPrefsR *PrefsR;
const struct sPrefsR DefaultPrefsR={
    PREFSVERSION,
    /* General Prefs */
    { "10", '1', '1', '0', { "mail", "date", "addr", "pref", "todo", "memo" } },
    /* SMTP Prefs */
    { "smtpserv", "25", "dick@weed.net", "Dick Weed", "", SMTPPrefs_FILED,
        0, "", "" },
    /* POP Prefs */
    { "popserv", "110", "user", "pass", "0", "1",
        POPPrefs_DETECTRESET|POPPrefs_DELETEMSG,
        "firstuidl" }
};

/*
 * These are for the private database
 */
DmOpenRef MainDB=0;
const char MainDBName[] = "gnugotmailDB\0";

/*
 * These are for accessing the built-in MailDB
 */
DmOpenRef MailDB=0;
const char MailDBName[] = "MailDB\0";
Int32 RecordCount;

/*
 * This migrates preferences from the last release to the current one.
 */
void MigratePrefs(struct sPrefsR *currP, struct s12PrefsR *oldP)
{
    char flags;
    /* first initialize a new structure to the defaults */
    MemMove(currP, &DefaultPrefsR, sizeof(DefaultPrefsR));
    /* then copy over stuff from the older prefs, where applicable */
    /* General */
    MemMove(&currP->Prefs, &oldP->Prefs, sizeof(currP->Prefs));
    /* SMTP */
    MemMove(&currP->SMTPPrefs.server, &oldP->SMTPPrefs.server,
        sizeof(oldP->SMTPPrefs.server));
    MemMove(&currP->SMTPPrefs.port, &oldP->SMTPPrefs.port,
        sizeof(oldP->SMTPPrefs.port));
    MemMove(&currP->SMTPPrefs.email, &oldP->SMTPPrefs.email,
        sizeof(oldP->SMTPPrefs.email));
    MemMove(&currP->SMTPPrefs.name, &oldP->SMTPPrefs.name,
        sizeof(oldP->SMTPPrefs.name));
    MemMove(&currP->SMTPPrefs.addbcc, &oldP->SMTPPrefs.addbcc,
        sizeof(oldP->SMTPPrefs.addbcc));
    /* convert to bitmap flags rather than '1' or '0' chars */
    flags = 0;
    if (oldP->SMTPPrefs.filed == '1') flags |= SMTPPrefs_FILED;
    if (oldP->SMTPPrefs.fakepop == '1') flags |= SMTPPrefs_FAKEPOP;
    currP->SMTPPrefs.flags = flags;
    /* POP */
    MemMove(&currP->POPPrefs.server, &oldP->POPPrefs.server,
        sizeof(oldP->POPPrefs.server));
    MemMove(&currP->POPPrefs.port, &oldP->POPPrefs.port,
        sizeof(oldP->POPPrefs.port));
    MemMove(&currP->POPPrefs.user, &oldP->POPPrefs.user,
        sizeof(oldP->POPPrefs.user));
    MemMove(&currP->POPPrefs.pass, &oldP->POPPrefs.pass,
        sizeof(oldP->POPPrefs.pass));
    MemMove(&currP->POPPrefs.maxlines, &oldP->POPPrefs.maxlines,
        sizeof(oldP->POPPrefs.maxlines));
    MemMove(&currP->POPPrefs.startmsg, &oldP->POPPrefs.startmsg,
        sizeof(oldP->POPPrefs.startmsg));
    /* convert to bitmap flags rather than '1' or '0' chars */
    flags = 0;
    if (oldP->POPPrefs.detectreset == '1') flags |= POPPrefs_DETECTRESET;
    if (oldP->POPPrefs.deletemsg == '1') flags |= POPPrefs_DELETEMSG;
    if (oldP->POPPrefs.useapop == '1') flags |= POPPrefs_USEAPOP;
}

Boolean MainFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    RectangleType r, r1;
    Boolean ishighlit, isinside, penDown;
    Int16 x, y;
    Int16 i;
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        ShowStatusLine(true);
        ShowIcons();
        handled = true;
        break;

    case frmUpdateEvent:
        /*
         * Not very likely, but the OS3.5 debug ROM was built to
         * provide this often for testing your program's reaction to low
         * memory conditions.  We return handled = true so that the default
         * action of the OS doesn't call FrmDrawForm() again.
         */
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        ShowStatusLine(true);
        ShowIcons();
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch(e->data.menu.itemID) {
        case mitemID_prefs:
            FrmGotoForm(PrefsForm);
            break;
        case mitemID_popprefs:
            FrmGotoForm(POPPrefsForm);
            break;
        case mitemID_smtpprefs:
            FrmGotoForm(SMTPPrefsForm);
            break;
        case mitemID_sendmail:
            /*
             * Sending mail can also involve sending a fake POP login.  If this
             * is enabled, we will call OpenPOP3() which will do a POP login
             * and then it will eventually cascade us into OpenSMTP().
             */
            if (PrefsR->SMTPPrefs.flags&SMTPPrefs_FAKEPOP) {
                if (!popsock && !smtpsock) OpenPOP3(2);
            } else {
                if (!smtpsock) OpenSMTP();
            }
            break;
        case mitemID_getmail:
            if (!popsock) OpenPOP3(0);
            break;
        case mitemID_highmsg:
            if (!popsock) OpenPOP3(1);
            break;
        case mitemID_about:
            FrmPopupForm(AboutForm);
            break;
        case mitemID_softreset:
            /*
             * I changed over to EVPlugBase from Hackmaster and I miss the
             * soft reset option.  So, well, here it is!
             */
            frm = FrmInitForm(SoftResetForm);
            i = FrmDoDialog(frm);
            FrmDeleteForm(frm);
            if (i == buttonID_softresetyes) SysReset();
            break;
        }
       handled = true;
       break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_sendmail:
            /*
             * Sending mail can also involve sending a fake POP login.  If this
             * is enabled, we will call OpenPOP3() which will do a POP login
             * and then it will eventually cascade us into OpenSMTP().
             */
            if (PrefsR->SMTPPrefs.flags&SMTPPrefs_FAKEPOP) {
                if (!popsock && !smtpsock) OpenPOP3(2);
            } else {
                if (!smtpsock) OpenSMTP();
            }
            break;
        case buttonID_getmail:
            if (!popsock) OpenPOP3(0);
            break;
        case buttonID_clearinbox:
            ClearInbox();
            break;
        case buttonID_cancel:
            ClosePOP3();
            CloseSMTP();
            ShowErrorMsg("user cancelled.");
            /* this global controls our return param */
            backgroundretval = RV_USERCANCEL;
            break;
        }
        break;

    case penDownEvent:
        /* first check if the icons are currently showing */
        if (buttonsshown != 0) break;
        frm = FrmGetActiveForm();
        /*
         * Checking if the user is pressing one of the six launch icons.  I
         * have gadgets to represent where they should be.
         */
        isinside = true;
        do {
        i = 0;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch1), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch2), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch3), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch4), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch5), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        i++;
        FrmGetObjectBounds(frm, FrmGetObjectIndex(frm, gadgetID_launch6), &r);
        if (RctPtInRectangle(e->screenX, e->screenY, &r)) break;
        isinside = false;
        } while (0);
        if (!isinside) break;
        ishighlit = false;
        r1 = r;
        r1.extent.y = 24;
        /*
         * While the pen is held down I stay in this loop.  The user can move
         * the stylus outside of the area to stop the operation.
         */
        do {
            PenGetPoint(&x, &y, &penDown);
            isinside = RctPtInRectangle(x, y, &r);
            if ((!ishighlit && isinside) ||
                (ishighlit && !isinside)) {
                WinInvertRectangle(&r1, 5);
                ishighlit = !ishighlit;
            }
        } while (penDown);
        /*
         * If the stylus was still inside the icon when released then launch
         * the application.
         */
        if (ishighlit) {
            WinInvertRectangle(&r1, 5);
            if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
                sysFileTApplication, Prefslaunch[i], true, &cardNo,
                &dbID) == 0)) {
                SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdNormalLaunch, NULL);
            }
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean PrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    UInt32 AppID;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadPrefsFields();
        FrmDrawForm(frm);
        FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_prefsleavefree));
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        handled = FormHandleMenuEvent(e->data.menu.itemID);
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case popuptriggerID_prefslaunch1:
            MemMove(&AppID, PrefsR->Prefs.launch[0], sizeof(AppID));
            Prefslaunch[0] = AppsPopup(listID_prefslaunch1,
                popuptriggerID_prefslaunch1, AppID);
            break;
        case popuptriggerID_prefslaunch2:
            MemMove(&AppID, PrefsR->Prefs.launch[1], sizeof(AppID));
            Prefslaunch[1] = AppsPopup(listID_prefslaunch2,
                popuptriggerID_prefslaunch2, AppID);
            break;
        case popuptriggerID_prefslaunch3:
            MemMove(&AppID, PrefsR->Prefs.launch[2], sizeof(AppID));
            Prefslaunch[2] = AppsPopup(listID_prefslaunch3,
                popuptriggerID_prefslaunch3, AppID);
            break;
        case popuptriggerID_prefslaunch4:
            MemMove(&AppID, PrefsR->Prefs.launch[3], sizeof(AppID));
            Prefslaunch[3] = AppsPopup(listID_prefslaunch4,
                popuptriggerID_prefslaunch4, AppID);
            break;
        case popuptriggerID_prefslaunch5:
            MemMove(&AppID, PrefsR->Prefs.launch[4], sizeof(AppID));
            Prefslaunch[4] = AppsPopup(listID_prefslaunch5,
                popuptriggerID_prefslaunch5, AppID);
            break;
        case popuptriggerID_prefslaunch6:
            MemMove(&AppID, PrefsR->Prefs.launch[5], sizeof(AppID));
            Prefslaunch[5] = AppsPopup(listID_prefslaunch6,
                popuptriggerID_prefslaunch6, AppID);
            break;
        case buttonID_prefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_prefscancel:
            FrmGotoForm(MainForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean SMTPPrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadSMTPPrefsFields();
        FrmDrawForm(frm);
        FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_smtpserver));
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadSMTPPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        handled = FormHandleMenuEvent(e->data.menu.itemID);
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_smtpprefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_smtpprefscancel:
            FrmGotoForm(MainForm);
            break;
        case triggerID_smtppass:
            FrmPopupForm(SMTPPrefsPassForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean SMTPPrefsPassFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    VoidPtr obj;
    char *ts;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_smtppass));
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch(e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_smtpprefspassok:
            frm = FrmGetActiveForm();
            obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_smtppass));
            MemSet(PrefsR->SMTPPrefs.passwd,
                sizeof(PrefsR->SMTPPrefs.passwd), 0);
            if ((ts = FldGetTextPtr(obj)))
                MemMove(PrefsR->SMTPPrefs.passwd, ts, StrLen(ts));
            FrmReturnToForm(0);
            handled = true;
            break;
        case buttonID_smtpprefspasscancel:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean POPPrefsFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    static Boolean save;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        LoadPOPPrefsFields();
        FrmDrawForm(frm);
        FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_popserver));
        save = false;
        handled = true;
        break;

    case frmCloseEvent:
        UnloadPOPPrefsFields(save);
        if (save) WritePrefsRec();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        handled = FormHandleMenuEvent(e->data.menu.itemID);
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_popprefsok:
            save = true;
            FrmGotoForm(MainForm);
            break;
        case buttonID_popprefscancel:
            FrmGotoForm(MainForm);
            break;
        case triggerID_poppass:
            FrmPopupForm(POPPrefsPassForm);
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean POPPrefsPassFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    VoidPtr obj;
    char *ts;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_poppass));
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch(e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_popprefspassok:
            frm = FrmGetActiveForm();
            obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                fieldID_poppass));
            MemSet(PrefsR->POPPrefs.pass,
                sizeof(PrefsR->POPPrefs.pass), 0);
            if ((ts = FldGetTextPtr(obj)))
                MemMove(PrefsR->POPPrefs.pass, ts, StrLen(ts));
            FrmReturnToForm(0);
            handled = true;
            break;
        case buttonID_popprefspasscancel:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean AboutFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_aboutok:
            FrmReturnToForm(0);
            handled = true;
            break;
        case buttonID_aboutcontrib:
            FrmPopupForm(ContribForm);
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean ContribFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    ListPtr lst;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listID_contrib));
        /* turn off the highlight */
        LstSetSelection(lst, -1);
        FrmDrawForm(frm);
        handled = true;
        break;

    case menuEvent:
        MenuEraseStatus(NULL);
        switch (e->data.menu.itemID) {
        }
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlSelect.controlID) {
        case buttonID_contribok:
            FrmReturnToForm(0);
            handled = true;
            break;
        }
        break;

    case keyDownEvent:
        if (e->data.keyDown.chr == pageUpChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winUp, LstGetVisibleItems(lst));
        } else if (e->data.keyDown.chr == pageDownChr) {
            frm = FrmGetActiveForm();
            lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
                listID_contrib));
            LstScrollList(lst, winDown, LstGetVisibleItems(lst));
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

Boolean BackgroundFormHandleEvent(EventPtr e)
{
    Boolean handled;
    FormPtr frm;

    CALLBACK_PROLOGUE

    handled = false;

    switch (e->eType) {
    case frmUpdateEvent:
        /*
         * Not very likely, but the OS3.5 debug ROM was built to
         * provide this often for testing your program's reaction to low
         * memory conditions.  We return handled = true so that the default
         * action of the OS doesn't call FrmDrawForm() again.
         */
        frm = FrmGetActiveForm();
        FrmDrawForm(frm);
        handled = true;
        break;

    case ctlSelectEvent:
        switch(e->data.ctlSelect.controlID) {
        case buttonID_backgroundcancel:
            ClosePOP3();
            CloseSMTP();
            ShowErrorMsg("user cancelled.");
            /* this global controls our return param */
            backgroundretval = RV_USERCANCEL;
            break;
        }
        break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

/*
 * Returns field that has the focus, if any, including in embedded tables.
 */
FieldPtr GetFocusObjectPtr(void)
{
    FormPtr frm;
    UInt16 focus;
    FormObjectKind objType;
    frm = FrmGetActiveForm();
    focus = FrmGetFocus(frm);
    if (focus == noFocus) return NULL;
    objType = FrmGetObjectType(frm, focus);
    if (objType == frmFieldObj) return FrmGetObjectPtr(frm, focus);
    else if (objType == frmTableObj)
        return TblGetCurrentField(FrmGetObjectPtr(frm, focus));
    return NULL;
}

Boolean FormHandleMenuEvent(UInt16 menuID)
{
    FieldPtr fld;
    switch (menuID) {
    case sysEditMenuUndoCmd:
    case sysEditMenuCutCmd:
    case sysEditMenuCopyCmd:
    case sysEditMenuPasteCmd:
    case sysEditMenuSelectAllCmd:
        fld = GetFocusObjectPtr();
        if (!fld) return false;
        if (menuID == sysEditMenuUndoCmd) FldUndo(fld);
        else if (menuID == sysEditMenuCutCmd) FldCut(fld);
        else if (menuID == sysEditMenuCopyCmd) FldCopy(fld);
        else if (menuID == sysEditMenuPasteCmd) FldPaste(fld);
        else if (menuID == sysEditMenuSelectAllCmd)
            FldSetSelection (fld, 0, FldGetTextLength (fld));
        return true;
    case sysEditMenuKeyboardCmd:
        SysKeyboardDialog(kbdDefault);
        return true;
    case sysEditMenuGraffitiCmd:
        SysGraffitiReferenceDialog(referenceDefault);
        return true;
    }
    return false;
}

Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    UInt16    formId;
    Boolean handled;

    CALLBACK_PROLOGUE

    handled = false;
    if (e->eType == frmLoadEvent) {
        formId = e->data.frmLoad.formID;
        frm = FrmInitForm(formId);
        FrmSetActiveForm(frm);

        switch(formId) {
        case MainForm:
            FrmSetEventHandler(frm, MainFormHandleEvent);
            break;
        case PrefsForm:
            FrmSetEventHandler(frm, PrefsFormHandleEvent);
            break;
        case POPPrefsForm:
            FrmSetEventHandler(frm, POPPrefsFormHandleEvent);
            break;
        case POPPrefsPassForm:
            FrmSetEventHandler(frm, POPPrefsPassFormHandleEvent);
            break;
        case SMTPPrefsForm:
            FrmSetEventHandler(frm, SMTPPrefsFormHandleEvent);
            break;
        case SMTPPrefsPassForm:
            FrmSetEventHandler(frm, SMTPPrefsPassFormHandleEvent);
            break; 
        case AboutForm:
            FrmSetEventHandler(frm, AboutFormHandleEvent);
            break;
        case ContribForm:
            FrmSetEventHandler(frm, ContribFormHandleEvent);
            break;
        case BackgroundForm:
            FrmSetEventHandler(frm, BackgroundFormHandleEvent);
            break;
        }
        handled = true;
    }

    CALLBACK_EPILOGUE

    return handled;
}

/*
 * These are called in StartApplication and StopApplication because they
 * are required memory allocations that exist from start to finish of the
 * application run.
 */
void freememories(void)
{
    if (popd.msgrec) MemPtrFree(popd.msgrec);
    if (popd.sq.sendq) MemPtrFree(popd.sq.sendq);
    if (smtpd.sq.sendq) MemPtrFree(smtpd.sq.sendq);
    if (PrefsRHandle) MemHandleFree(PrefsRHandle);
}

/* Get preferences, open (or create) app database */
UInt16 StartApplication(void)
{
    Err err;
    Boolean loaded=false;

    tps = SysTicksPerSecond();
    AppNetTimeout = 60*tps;
    AppNetRefnum = -1;
    buttonsshown = 0;
    popd.msgrec = 0;
    popd.sq.sendq = 0;
    popd.sq.sendqin = 0;
    popd.sq.sendqout = 0;
    smtpd.sq.sendq = 0;
    smtpd.sq.sendqin = 0;
    smtpd.sq.sendqout = 0;
    PrefsRHandle = 0;
    PrefsR = NULL;
    PrefsRdirty = false;
    PrefsWeredirty = false;
    if ((popd.msgrec = MemPtrNew(MSGRECSIZE)) == 0) {
        freememories();
        return 1;
    }
    if ((popd.sq.sendq = MemPtrNew(SENDQSIZE)) == 0) {
        freememories();
        return 1;
    }
    if ((smtpd.sq.sendq = MemPtrNew(SENDQSIZE)) == 0) {
        freememories();
        return 1;
    }
    if ((PrefsRHandle = MemHandleNew(sizeof(struct sPrefsR))) == 0) {
        freememories();
        return 1;
    }
    /*
     * Lock the PrefsR structure
     */
    PrefsR = MemHandleLock(PrefsRHandle);
    if (!OpenDatabases()) {
        freememories();
        return 1;
    }

    if (!isbackground) {
        FrmGotoForm(MainForm);
    } else {
    }

    /* load/open ZLib */
    if (!ZLibRef) {
        /* try to find it */
        err = SysLibFind("Z.lib", &ZLibRef);
        if (err != errNone) {
            /* try to load it */
            err = SysLibLoad('libr', 'ZLib', &ZLibRef);
            if (err == errNone) loaded = true;
        }
        err = ZLibOpen(ZLibRef);
        if (err != errNone) {
            if (loaded) SysLibRemove(ZLibRef);
            ZLibRef = 0;
        }
    }

    return 0;
}

/* Save preferences, close forms, close app database */
void StopApplication(void)
{
    UInt16 zltmp;
    if (!isbackground) {
        FrmSaveAllForms();
        FrmCloseAllForms();
    }
    ClosePOP3();
    CloseSMTP();
    CloseDatabases();
    freememories();
    /* close/unload ZLib */
    if (ZLibRef) {
        ZLibClose(ZLibRef, &zltmp);
        if (!zltmp) SysLibRemove(ZLibRef);
        ZLibRef = 0;
    }
}

/* The main event loop */
void EventLoop(void)
{
    Err err;
    EventType e;
    Int16 rv, bytes;
    NetFDSetType rfds, wfds;
    Int32 l1; 
    Int32 eventTimeout, to, lastticks=0;
    Boolean onmainform=false;
    int k;
    unsigned char *p;

    lastticks = TimGetTicks();
    do {

    if (popsock || smtpsock || (smtpd.parseState == smtpstCLEARBITS)) {
        /*
         * This select() is for all open sockets, and also it will
         * come out if there are any pending events which lets the
         * normal stuff happen.
         */
        netFDZero(&rfds);
        netFDZero(&wfds);
        netFDSet(sysFileDescStdIn, &rfds);
        /*
         * This tries to make sure that reads won't occur if there
         * is data still enqueued to go out.  Not very efficient, but
         * necessary so that I don't enqueue too much outgoing data.
         */
        if (popsock && !popd.sq.sendqin) netFDSet(popsock, &rfds);
        if (smtpsock && !smtpd.sq.sendqin) netFDSet(smtpsock, &rfds);
        /*
         * If there is any data enqueued for a connection we will set
         * the bit here so that it will indicate to us when it is ready
         * for writing.
         */
        if (popsock && popd.sq.sendqin) netFDSet(popsock, &wfds);
        if (smtpsock && smtpd.sq.sendqin) netFDSet(smtpsock, &wfds);
        /*
         * This select will drop out; when incoming data arrives, when
         * outgoing data may be sent, when an OS event occurs, and when
         * the timeout expires.  I set this to a reasonable value, unless
         * the smtptick() is in the CLEARBITS state because then I want
         * it to go very fast.
         *
         * We also do this during BODY2 so that the user may cancel during
         * file enumeration.  Normally we return here once some outgoing
         * data goes, but in BODY2 state we only output data when we find
         * an appropriate file.  So unless we allow this timeout, the user
         * wouldn't be able to cancel because I would need to stay inside
         * smtptick().
         */
        if (((smtpd.parseState == smtpstCLEARBITS) ||
            (!smtpd.sq.sendqin && (smtpd.parseState == smtpstBODY2)))) to = 0;
        else to = 2*tps;
        if (isnetlibopen) {
            rv = NetLibSelect(AppNetRefnum, netFDSetSize, &rfds, &wfds, NULL,
                to, &errno);
        } else {
            rv = 0;
        }
        if (onmainform) {
            /*
             * Update the status line whenever a timeout occurs.
             */
            to = TimGetTicks();
            if ((to-lastticks) >= (2*tps)) {
                ShowStatusLine(false);
                lastticks = to;
            }
        }
        if (rv > 0) {
            /*
             * Incoming data
             */
            if (popsock && netFDIsSet(popsock, &rfds)) {
                if ((bytes = NetLibReceive(AppNetRefnum, popsock, ibuf,
                    sizeof(ibuf), 0, 0, 0, AppNetTimeout, &errno)) > 0) {
                    /*
                     * Keep the unit on during network activity.
                     */
                    if (PrefsR->Prefs.noautooff == '1') EvtResetAutoOffTimer();
                    /* Here we inflate a compressed stream */
                    p = ibuf;
                    parse_data_again:
                    l1 = 0;
                    if (popd.isgzip) {
                        pgpz.next_in = p;
                        pgpz.avail_in = bytes;
                        while (pgpz.avail_in) {
                            pgpz.next_out = obuf;
                            pgpz.avail_out = sizeof(obuf);
                            k = inflate(&pgpz, 0);
                            l1 = poptick(&popd, obuf,
                                sizeof(obuf)-pgpz.avail_out, popsock);
                            if (k != Z_OK) {
                                ClosePOP3();
                                break;
                            }
                        }
                    } else {
                        l1 = poptick(&popd, ibuf, bytes, popsock);
                        /*
                         * This condition means that I need to recheck
                         * whether we've switched into compressed mode,
                         * and then send the current buffer back into
                         * poptick().
                         */
                        if ((l1 > 0) && (l1 < bytes)) {
                            p = &ibuf[l1];
                            bytes -= l1;
                            goto parse_data_again;
                        }
                    }
                    if (l1 < 0) ClosePOP3();
                } else ClosePOP3();
            }
            if (smtpsock && netFDIsSet(smtpsock, &rfds)) {
                if ((bytes = NetLibReceive(AppNetRefnum, smtpsock, ibuf,
                    sizeof(ibuf), 0, 0, 0, AppNetTimeout, &errno)) > 0) {
                    /*
                     * Keep the unit on during network activity.
                     */
                    if (PrefsR->Prefs.noautooff == '1') EvtResetAutoOffTimer();
                    l1 = smtptick(ibuf, bytes, smtpsock);
                    if (l1 < 0) CloseSMTP();
                } else CloseSMTP();
            }
            /*
             * Checking for channels ready for writing.  These will only
             * be true if there is data in the sendq because of the
             * netFDSet() above.
             */
            if (popsock && netFDIsSet(popsock, &wfds)) {
                if ((bytes = NetLibSend(AppNetRefnum, popsock,
                    &popd.sq.sendq[popd.sq.sendqout],
                    popd.sq.sendqin-popd.sq.sendqout, 0, 0, 0, AppNetTimeout,
                    &errno)) > 0) {
                    /*
                     * Keep the unit on during network activity.
                     */
                    if (PrefsR->Prefs.noautooff == '1') EvtResetAutoOffTimer();
                    popd.sq.sendqout += bytes;
                    if (popd.sq.sendqin == popd.sq.sendqout) {
                        popd.sq.sendqin = 0;
                        popd.sq.sendqout = 0;
                    }
                } else {
                    ClosePOP3();
                }
            }
            if (smtpsock && netFDIsSet(smtpsock, &wfds)) {
                if ((bytes = NetLibSend(AppNetRefnum, smtpsock,
                    &smtpd.sq.sendq[smtpd.sq.sendqout],
                    smtpd.sq.sendqin-smtpd.sq.sendqout, 0, 0, 0, AppNetTimeout,
                    &errno)) > 0) {
                    /*
                     * Keep the unit on during network activity.
                     */
                    if (PrefsR->Prefs.noautooff == '1') EvtResetAutoOffTimer();
                    smtpd.sq.sendqout += bytes;
                    if (smtpd.sq.sendqin == smtpd.sq.sendqout) {
                        smtpd.sq.sendqin = 0;
                        smtpd.sq.sendqout = 0;
                        /*
                         * This needs a little explanation.  At some time I
                         * will be needing to enqueue more data than will fit
                         * in the sendq.  For example the BODY state.  So I
                         * give the parser a call when the sendq has finished
                         * flushing.  Note that I call this with a length of
                         * zero because it is a special case.
                         */
                        smtptick(&nullchar, 0, smtpsock);
                    }
                } else {
                    CloseSMTP();
                }
            }
        } else {
            /*
             * The CLEARBITS state wants me to call smtptick() as fast as
             * I can.  So when in that state I make the select timeout = 0.
             * The purpose of this is so that I can iterate through files
             * and still be responsive to the user's cancel request.
             */
            if (((smtpd.parseState == smtpstCLEARBITS) ||
                (!smtpd.sq.sendqin && (smtpd.parseState == smtpstBODY2)))) {
                smtptick(&nullchar, 0, smtpsock);
                /*
                 * CLEARBITS will throw me into this state if the Clear
                 * Bits button was pressed.
                 */
                if (!smtpsock && (smtpd.parseState == smtpstALLDONE))
                    CloseSMTP();
            }
        }
        eventTimeout = 0;
    } else eventTimeout = 2*tps;
    
    EvtGetEvent(&e, eventTimeout);
    /*
     * If one of the silk screened keys is hit we want
     * to shutdown the connection.
     */
    if ((e.eType == keyDownEvent) &&
        (e.data.keyDown.modifiers & commandKeyMask) &&
        ((e.data.keyDown.chr == launchChr) ||
        (e.data.keyDown.chr == menuChr) ||
        (e.data.keyDown.chr == autoOffChr) ||
        (e.data.keyDown.chr == hardPowerChr))) {
        ClosePOP3();
        CloseSMTP();
    }
    /*
     * Update the status line whenever a nilEvent occurs.  The showstatus
     * variable is used to determine if there is another form loaded.. in
     * which case we don't want to obscure it with the status line.
     */
    if (onmainform && (e.eType == nilEvent) &&
        !(popsock || smtpsock)) {
        ShowStatusLine(false);
    }

    /* Auto-start a POP.  This is enqueued by subcall. */
    if (onmainform && !popsock && (e.eType == EV_CHECKMAIL)) {
        if (OpenPOP3(0) && !isbackground) {
            EvtSetAutoOffTimer(SetExactly, 5); /* power off in 5 seconds */
        }
    }

    /* Auto-start an SMTP.  This is enqueued by subcall. */
    if (onmainform && !smtpsock && (e.eType == EV_SENDMAIL)) {
        if (!OpenSMTP() && !isbackground) {
            EvtSetAutoOffTimer(SetExactly, 5); /* power off in 5 seconds */
        }
    }

    if (!SysHandleEvent(&e)) {
        if (!MenuHandleEvent(NULL, &e, &err)) {
            if (!ApplicationHandleEvent(&e)) {
                FrmDispatchEvent(&e);
            }
        }
    }

    if (e.eType == winEnterEvent) {
        if (e.data.winEnter.enterWindow ==
            (WinHandle) FrmGetFormPtr(MainForm))
            onmainform = true; else onmainform = false;
    }

    } while ((e.eType != appStopEvent) && (e.eType != EV_STOPLOOP));

    if (!isbackground && (e.eType == EV_STOPLOOP)) {
        EvtSetAutoOffTimer(SetExactly, 5); /* power off in 5 seconds */
    }

}

/* From Handspring's PhoneLib Ref.pdf */

UInt32 WhichPhoneType(void)
{
    UInt32 phoneType;

    /* Check what device we are on: GSM or CDMA */
    (void)HsAttrGet(hsAttrPhoneType, 0, &phoneType);
    switch (phoneType) {
    case hsAttrPhoneTypeGSM:
    case hsAttrPhoneTypeCDMA:
        break;
    default:
        /* Pre Treo 300 phones should default to GSM */
        phoneType = hsAttrPhoneTypeGSM;
        break;
    }
    return phoneType;
}

Err LoadPhoneLibrary(UInt16 *plref)
{
    Err error=0;

    *plref = -1;

    switch (WhichPhoneType()) {
    case hsAttrPhoneTypeGSM:
        /* Get the Phone library */
        error = SysLibFind(phnLibName, plref);
        if (error) {
            /* It's not already here - have to load it ourselves */
            error = SysLibLoad(phnLibDbType, phnLibDbCreator, plref);
        }
        break;
    case hsAttrPhoneTypeCDMA:
        /* Get the Phone library */
        error = SysLibFind("PhoneLib.prc", plref);
        if (error) {
            /* It's not already here - have to load it ourselves */
            error = SysLibLoad(phnLibDbType, 'HsCL', plref);
        }
        break;
    }

    return error;
}

void ClosePhoneLibrary(UInt16 plref)
{
    if (plref != -1) {
        PhnLibClose(plref);
    }
}

Boolean is_ok_to_connect(void)
{
    UInt32 tui32;
    UInt16 tplibref;
    Boolean retval=true;

    if (!FtrGet(hsFtrCreator, hsFtrIDVersion, &tui32)) {
        /* handspring extensions, so check if phone is in use */
        if (!LoadPhoneLibrary(&tplibref)) {
            /*
             * Let's make sure the user isn't in the middle of a
             * phone call.
             */
            if (!PhnLibGetPhoneCallStatus(tplibref, &tui32) &&
                (tui32&(phnVoiceCall1Active|phnVoiceCall2Active))) {
                retval = false;
            }
            ClosePhoneLibrary(tplibref);
        }
    }
    return retval;
}

Err OpenPOP3(Int16 whattodo)
{
    Err err, ifErrs;
    char buf[46+ /* sizes of user=&pass=&top=&start=&cnt=#####&del=1&uidl1= */
        sizeof(PrefsR->POPPrefs.user)+sizeof(PrefsR->POPPrefs.pass)+
        sizeof(PrefsR->POPPrefs.maxlines)+sizeof(PrefsR->POPPrefs.startmsg)];
    char buf2[]="Content-Length: ####\r\n\r\n";
    UInt16 cnt=65535;

    err = SysLibFind("Net.lib", &AppNetRefnum);
    ErrNonFatalDisplayIf(err, ErrToString(err));
    /* don't do it if a connection isn't already made */
    if (!is_ok_to_connect()) return RV_FAILED;
    ShowInfoMsg("");
    ShowButtons(1);
    /*
     * This alarm is used to click-through any stupid dialogs that
     * stay on the screen waiting for "OK".
     */
    hit_ok_dialog();
    err = NetLibOpen(AppNetRefnum, &ifErrs);
    if ((err && err != netErrAlreadyOpen) || ifErrs) {
        /*
         * We always want to use a true on the NetLibClose() if there
         * was a failure with NetLibOpen().  Otherwise the PalmOS goes
         * berzerk thinking that it has a network connection when it
         * doesn't.. and it makes you cycle power to fix.
         */
        if (NetLibClose(AppNetRefnum, true) != netErrStillOpen) {
            isnetlibopen = false;
            ShowButtons(0);
        }
        //NetLibFinishCloseWait(AppNetRefnum);
        ShowErrorMsg(ErrToString(ifErrs));
        return ifErrs;
    }
    isnetlibopen = true;
    /*
     * Here is where I would open the connection to the
     * POP3 server, which should start the train rolling.
     */
    popd.error = 0;
    popd.parseAction = -1;
    popd.keepdata = 0;
    popd.skipdata = 0;
    popd.msgreccnt = 0;
    popd.startmsg = StrAToI(PrefsR->POPPrefs.startmsg);
    popd.msgnum = popd.startmsg;
    popd.maxlines = StrAToI(PrefsR->POPPrefs.maxlines);
    leavefree = StrAToI(PrefsR->Prefs.leavefree);
    popd.msgrechandle = 0;
    popd.sq.sendqin = 0;
    popd.sq.sendqout = 0;
    popd.justhighmsg = false;
    popd.onlylogin = false;
    popd.wasreset = false;
    popd.isgzip = false;
    if (PrefsR->POPPrefs.flags&POPPrefs_USEHTTP) {
        popd.usehttp = true;
        popd.parseState = stHTTPSTART;
    } else {
        popd.usehttp = false;
        popd.parseState = stOPEN;
    }
    switch (whattodo) {
    case 1:
        popd.justhighmsg = true;
        cnt = 0; /* for httpop don't bother getting any messages */
        break;
    case 2:
        popd.onlylogin = true;
        cnt = 0; /* for httpop don't bother getting any messages */
        break;
    }
    /*
     * Clear out the history chars
     */
    lastchars[0] = (char) NULL;
    lastchars[1] = (char) NULL;
    lastchars[2] = (char) NULL;
    lastchars[3] = (char) NULL;
    lastchars[4] = (char) NULL;
    lastchars[5] = (char) NULL;
    if ((popsock =
        connectsock(PrefsR->POPPrefs.server, PrefsR->POPPrefs.port, false))
        == 0) {
        if (NetLibClose(AppNetRefnum, false) != netErrStillOpen)
            isnetlibopen = false;
        //NetLibFinishCloseWait(AppNetRefnum);
        ShowButtons(0);
        ShowErrorMsg(NULL);
        return RV_FAILED;
    } else ShowButtons(2);
    /* start us off if we are using http */
    if (popd.usehttp) {
        writeall(popsock, &popd.sq,
            "POST /httpop.php HTTP/1.0\r\n", 27);
        StrPrintF(buf, "user=%s&pass=%s&top=%s&start=%s&cnt=%u&uidl1=%s%s",
            PrefsR->POPPrefs.user, PrefsR->POPPrefs.pass,
            PrefsR->POPPrefs.maxlines, PrefsR->POPPrefs.startmsg, cnt,
            PrefsR->POPPrefs.firstuidl,
            (PrefsR->POPPrefs.flags&POPPrefs_DELETEMSG)?"&del=1":"");
        if (ZLibRef) {
            /* yeah!  we support gzipped streams */
            writeall(popsock, &popd.sq,
                "Accept-Encoding: gzip\r\n", 23);
        }
        writeall(popsock, &popd.sq,
            "Content-type: application/x-www-form-urlencoded\r\n", 49);
        StrPrintF(buf2, "Content-length: %d\r\n\r\n", StrLen(buf));
        writeall(popsock, &popd.sq, buf2, StrLen(buf2));
        writeall(popsock, &popd.sq, buf, StrLen(buf));
    }

    /* get ZLib ready for possible use */
    if (ZLibRef) {
        memset(&pgpz, 0, sizeof(pgpz));
        inflateInit2(&pgpz, -15);
    }

    return RV_SUCCESS;
}

void ClosePOP3(void)
{
    int i1;

    if (!popsock) return;

    /* end ZLib usage */
    if (ZLibRef) {
        inflateEnd(&pgpz);
    }

    /*
     * It wasn't finished yet, so we'll make it a special error.
     */
    if ((popd.parseState != stALLDONE) && (!popd.error)) popd.error = 2;
    /* indicate a failed or partial retrieve -- it will likely reschedule */
    if (isbackground && popd.error) backgroundretval = RV_FAILED;
//    if (popd.usehttp) {
//        if (popd.parseAction == actGETBODY) {
//            /* allow it to terminate the final message */
//            poptick(&popd, &nullchar, 1, popsock);
//        }
//        popd.error = 0;
//    }
    if (popsock) {
        ShowButtons(3);
        popd.parseAction = -1;
        popd.parseState = stALLDONE;
        poptick(&popd, &nullchar, 1, popsock);
        /*
         * Keep the last message number if you are leaving mail on server.
         * But only if the 'Delete messages from server' is not set.  As you
         * recall this is meant to act as a way to leave your messages on
         * the server and just start reading at a particular message number.
         *
         * In addition we don't want to do this if this was the fake POP login
         * required by some SMTP servers before sending mail.
         */
        if ((!popd.onlylogin && !(PrefsR->POPPrefs.flags&POPPrefs_DELETEMSG)) ||
            popd.justhighmsg || popd.wasreset || popd.usehttp) {
            if (popd.wasreset) {
                /* UIDL mailbox reset stuff - let's start at msg #1 next time */
                tnum[0] = '1';
                tnum[1] = (char)NULL;
            } else {
                i1 = numtotxt(popd.msgnum, tnum, false);
            }
            MemSet(PrefsR->POPPrefs.startmsg,
                sizeof(PrefsR->POPPrefs.startmsg), 0);
            StrCopy(PrefsR->POPPrefs.startmsg, tnum);
            PrefsRdirty = true;
        }
        if (popsock)
            NetLibSocketClose(AppNetRefnum, popsock, AppNetTimeout, &errno);
        popsock = 0;
        /*
         * Here we are nice and use a false on the NetLibClose() to allow it
         * to stay open if others are using it.
         */
        if (isnetlibopen) {
            if (NetLibClose(AppNetRefnum, false) != netErrStillOpen)
                isnetlibopen = false;
            //NetLibFinishCloseWait(AppNetRefnum);
        }
        /*
         * We aren't done yet if we only did this POP to allow the user
         * to send mail via SMTP.  If that is the case, then we want to
         * start sending the mail now.
         */
        if (popd.onlylogin && !popd.error && !smtpsock) {
            OpenSMTP();
        } else {
            ShowButtons(0);
            if (popd.error == 1) ShowErrorMsg(NULL);
        }
    }
    if (isbackground) {
        post_custom_event(EV_STOPLOOP);
    }
}

Boolean OpenSMTP(void)
{
    Err err, ifErrs;
    /*
     * Why continue if there are no Outbox records.  This OutboxCount is
     * also used later. 
     */
    if ((OutboxCount = DmNumRecordsInCategory(MailDB, catOutbox)) <= 0)
        return false;
    ShowInfoMsg("");
    err = SysLibFind("Net.lib", &AppNetRefnum);
    ErrNonFatalDisplayIf(err, ErrToString(err));
    ShowButtons(1);
    /*
     * This alarm is used to click-through any stupid dialogs that
     * stay on the screen waiting for "OK".
     */
    hit_ok_dialog();
    err = NetLibOpen(AppNetRefnum, &ifErrs);
    if ((err && err != netErrAlreadyOpen) || ifErrs) {
        /*
         * We always want to use a true on the NetLibClose() if there
         * was a failure with NetLibOpen().  Otherwise the PalmOS goes
         * berzerk thinking that it has a network connection when it
         * doesn't.. and it makes you cycle power to fix.
         */
        if (NetLibClose(AppNetRefnum, true) != netErrStillOpen) {
            isnetlibopen = false;
            ShowButtons(0);
        }
        //NetLibFinishCloseWait(AppNetRefnum);
        ShowErrorMsg(ErrToString(ifErrs));
        return false;
    }
    isnetlibopen = true;
    /*
     * Here is where I open the connection to the SMTP server, which should
     * start the train rolling.  I initialize some critical variables here
     * because the smtptick() may not get called at all until CloseSMTP()
     * and I need to make sure that the variables that are tested in the
     * smtpstALLDONE state are valid.
     */
    smtpd.error = 0;
    smtpd.parseAction = -1;
    smtpd.parseState = smtpstOPEN;
    smtpd.keepdata = 0;
    smtpd.msgrechandle = 0;
    smtpd.sq.sendqin = 0;
    smtpd.sq.sendqout = 0;
    if ((smtpsock =
        connectsock(PrefsR->SMTPPrefs.server, PrefsR->SMTPPrefs.port, false))
        == 0) {
        if (NetLibClose(AppNetRefnum, false) != netErrStillOpen)
            isnetlibopen = false;
        //NetLibFinishCloseWait(AppNetRefnum);
        ShowButtons(0);
        ShowErrorMsg(NULL);
        return false;
    } else ShowButtons(2);
    return true;
}

void CloseSMTP(void)
{
    if (!smtpsock) return;
    /*
     * It wasn't finished yet, so we'll make it a special error.
     */
    if ((smtpd.parseState != smtpstALLDONE) && (!smtpd.error)) smtpd.error = 2;
    smtpd.parseAction = -1;
    smtpd.parseState = smtpstALLDONE;
    if (smtpsock) {
        ShowButtons(3);
        smtptick(&nullchar, 1, smtpsock);
        if (smtpsock)
            NetLibSocketClose(AppNetRefnum, smtpsock, AppNetTimeout, &errno);
        smtpsock = 0;
        /*
         * Here we are nice and use a false on the NetLibClose() to allow it
         * to stay open if others are using it.
         */
        if (isnetlibopen) {
            if (NetLibClose(AppNetRefnum, false) != netErrStillOpen)
                isnetlibopen = false;
            //NetLibFinishCloseWait(AppNetRefnum);
        }
        ShowButtons(0);
        if (smtpd.error == 1) ShowErrorMsg(NULL);
    }
    if (isbackground) {
        post_custom_event(EV_STOPLOOP);
    }
}

Boolean OpenDatabases(void)
{
    Err err;
    Ptr p, p2;
    Boolean migrated=false;

    /*
     * Set these to zero in case any of the opens fail.
     */
    PrefsRecHandle = 0;
    /*
     * Open/Create the private database
     */
    MainDB = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
        dmModeReadWrite);
    if (!MainDB) {
        /*
         * Create our database if it doesn't exist yet
         */
        if (DmCreateDatabase(0, MainDBName, MainAppID, MainDBType, false))
            return false;
        MainDB = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
            dmModeReadWrite);
        if (!MainDB) return false;
    }
    PrefsRecIndex = 0;
    PrefsRecHandle = DmQueryRecord(MainDB, PrefsRecIndex);
    if (PrefsRecHandle != 0) {
        /*
         * A Prefs record already exists so we will do a sanity check.
         */
        if (MemHandleSize(PrefsRecHandle) >= sizeof(PrefsR->version)) {
            /*
             * Let's compare the version number.  If it doesn't match we will
             * delete the record and make the user start from scratch.  Ha ha!
             *
             * BTW, the current Prefs record version is PREFSVERSION.
             * Change this each time the Prefs record layout changes.
             */
            p = MemHandleLock(PrefsRecHandle);
            if (*(char *)p != (char)PREFSVERSION) {
                if (*(char *)p == (char)(PREFSVERSION-1)) {
                    MigratePrefs(PrefsR, (struct s12PrefsR *) p);
                    migrated = true;
                }
                MemPtrUnlock(p);
                err = DmRemoveRecord(MainDB, PrefsRecIndex);
                ErrFatalDisplayIf(err, ErrToString(err));
                PrefsRecHandle = 0;
            } else MemPtrUnlock(p);
        } else {
            err = DmRemoveRecord(MainDB, PrefsRecIndex);
            ErrFatalDisplayIf(err, ErrToString(err));
            PrefsRecHandle = 0;
        }
    }
    if (PrefsRecHandle == 0) {
        /*
         * Create the preferences record if it doesn't exist and
         * preload it with default values.
         */
        PrefsRecHandle = DmNewRecord(MainDB, &PrefsRecIndex,
            sizeof(struct sPrefsR));
        ErrFatalDisplayIf(!PrefsRecHandle, "Could not create prefs record.");
        p = MemHandleLock(PrefsRecHandle);
        if (migrated) p2 = (Ptr) PrefsR; else p2 = (Ptr) &DefaultPrefsR;
        err = DmWrite(p, 0, p2, sizeof(struct sPrefsR));
        ErrFatalDisplayIf(err, "Could not write default prefs.");
        MemPtrUnlock(p);
        ReadPrefsRec();
        BuildIcons(); /* cause it to rebuild the launch icon records */
        DmReleaseRecord(MainDB, PrefsRecIndex, true);
    } else {
        ReadPrefsRec();
    }

    /*
     * A cheezy way to see if the launch icon cache records need to be rebuilt.
     */
    if (DmNumRecords(MainDB) < 7) BuildIcons();

    /* may already be set from Mail app during custom launch */
    if (!MailDB) MailDB = DmOpenDatabaseByTypeCreator(MailDBType, MailAppID,
        dmModeReadWrite | dmModeShowSecret);
    if (!MailDB) return false;
    popd.msgrechandle = 0;
    /*
     * Get the category indexes.  We don't use the category names any more
     * because it has come to my attention that they are spelled differently
     * on International versions.
     */

/*
    catInbox = CategoryFind(MailDB, "Inbox");
    catOutbox = CategoryFind(MailDB, "Outbox");
    catFiled = CategoryFind(MailDB, "Filed");
*/

    catInbox = 0;
    catOutbox = 1;
    catFiled = 3;

    RecordCount = DmNumRecordsInCategory(MailDB, catInbox);
    return true;
}

void CloseDatabases(void)
{
    UInt16 attr;

    if (MailDB) {
        endnewrecord();
        if (ismymaildb) DmCloseDatabase(MailDB);
    }
    if (MainDB) {
        if (PrefsRecHandle) {
            if (PrefsRdirty) WritePrefsRec();
            if (PrefsWeredirty) {
                DmRecordInfo(MainDB, PrefsRecIndex, &attr, NULL, NULL);
                attr |= dmRecAttrDirty;
                DmSetRecordInfo(MainDB, PrefsRecIndex, &attr, NULL);
            }
            /*
             * Unlock the PrefsR structure
             */
            if (PrefsR) MemPtrUnlock(PrefsR);
        }
        DmCloseDatabase(MainDB);
    }
}

void ClearInbox(void)
{
    UInt16 totalItems, i1, recordNum=0;
    MemHandle recordH;
    if (MailDB) {
        /*
         * Aha, this is from:
         *
         * Palm Programming: The Developer's Guide
         *
         * A book which I hope to get as soon as it shows up in my
         * bookstore.  Thankfully there is an online version for us
         * living in areas that don't have quality books.
         */
        if ((totalItems = DmNumRecordsInCategory(MailDB, catInbox)) <= 0)
            return; 
        recordNum = 0;
        ShowProgressBar(0, 1);
        for (i1=0; i1 < totalItems; i1++) {
            recordH = DmQueryNextInCategory(MailDB, &recordNum,
                catInbox);
            /*
             * Remove the record -- all of it.  Break if error.
             */
            if (DmRemoveRecord(MailDB, recordNum)) break;
            ShowProgressBar(i1, totalItems);
        }
        RecordCount = DmNumRecordsInCategory(MailDB, catInbox);
        ShowProgressBar(0, 0);
        ShowStatusLine(false);
    }
}

void AppsPopupDrawItem(UInt16 itemNum, RectanglePtr bounds,
    char **unusedP)
{
    SysDBListItemType *listP;
    char name[dmDBNameLength];
    UInt32 creator;
    Int16 tl, tw;
    Boolean ignored;
    char ellipsisChar=chrEllipsis;

    CALLBACK_PROLOGUE

    if (AppsPopupH != NULL) {
        listP = MemHandleLock((MemHandle) AppsPopupH);
        if (DmDatabaseInfo(listP[itemNum].cardNo, listP[itemNum].dbID,
            name, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
            &creator) == 0) {
            tl = StrLen(name);
            tw = FntCharsWidth(name, tl);
            if (tw <= bounds->extent.x) {
                WinDrawChars(name, StrLen(name), bounds->topLeft.x,
                    bounds->topLeft.y);
            } else {
                tw = bounds->extent.x-FntCharWidth(ellipsisChar);
                FntCharsInWidth(name, &tw, &tl, &ignored);
                WinDrawChars(name, tl, bounds->topLeft.x,
                    bounds->topLeft.y);
                WinDrawChars(&ellipsisChar, 1, bounds->topLeft.x+tw,
                    bounds->topLeft.y);
            }
        }
        MemPtrUnlock(listP);
    }

    CALLBACK_EPILOGUE
}

UInt32 AppsPopup(UInt16 listid, UInt16 triggerid, UInt32 curcreator)
{
    UInt16 error;
    Int curSelection=-1;
    Int newSelection;
    Int i1, width, height;
    UInt32 newcreator=curcreator;
    ListPtr lst;
    ControlPtr trig;
	FormPtr frm;
    char *label;
    RectangleType r;
    SysDBListItemType *listP;
	char name[dmDBNameLength];
    UInt32 creator;
    WinHandle saved=0;
    MemHandle strH;
    char *strP;

    /*
     * This is going to take a while, so we'll put a little message up
     * for the user.
     */
    if ((strH = DmGetResource(strRsc, strID_pleasewait))) {
        strP = MemHandleLock(strH);
        FntSetFont(boldFont);
        i1 = StrLen(strP);
        width = FntCharsWidth(strP, i1);
        height = FntCharHeight();
        r.topLeft.x = 80-((width+12)/2);
        r.topLeft.y = 80-((height+8)/2);
        r.extent.x = width+12;
        r.extent.y = height+8;
        if ((saved = WinSaveBits(&r, &error))) {
            WinEraseRectangle(&r, 0);
            r.topLeft.x++;
            r.topLeft.y++;
            r.extent.x -= 3;
            r.extent.y -= 3;
            WinDrawRectangleFrame(popupFrame, &r);
            WinDrawChars(strP, i1, r.topLeft.x+(r.extent.x/2)-(width/2),
                r.topLeft.y+(r.extent.y/2)-(height/2));
        }
        MemHandleUnlock(strH);
        DmReleaseResource(strH);
        FntSetFont(stdFont);
    }

    if (!SysCreateDataBaseList(sysFileTApplication, 0, &AppsPopupCount,
        &AppsPopupH, true)) {
        AppsPopupCount = 0;
        AppsPopupH = 0;
    } else {
        /* Figure out the current selection by comparing the creator ID's */
        listP = MemHandleLock((MemHandle) AppsPopupH);
        for (i1=0; i1 < AppsPopupCount; i1++) {
            if (listP[i1].creator == curcreator) {
                curSelection = i1;
                break;
            }
        }
        MemPtrUnlock(listP);
    }

    /* Get rid of the message */
    if (saved) WinRestoreBits(saved, --r.topLeft.x, --r.topLeft.y);

    if (AppsPopupCount) {
        frm = FrmGetActiveForm();
        lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, listid));
        LstSetDrawFunction(lst, (ListDrawDataFuncPtr) AppsPopupDrawItem);

        LstSetListChoices(lst, NULL, AppsPopupCount);
        LstSetSelection(lst, curSelection);
        LstSetTopItem(lst, curSelection);

        newSelection = LstPopupList(lst);

        if ((newSelection != curSelection) && (newSelection >= 0)) {
            listP = MemHandleLock((MemHandle) AppsPopupH);
            if (DmDatabaseInfo(listP[newSelection].cardNo,
                listP[newSelection].dbID, name, NULL, NULL, NULL, NULL,
                NULL, NULL, NULL, NULL, NULL, &creator) == 0) {
                newcreator = creator;
                /*
                 * Copy the name into the popup trigger label.
                 */
                trig = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, triggerid));
                label = (char *)CtlGetLabel(trig);
                StrCopy(label, name);
                CtlSetLabel(trig, label);
            }
            MemPtrUnlock(listP);
        }
    }
    
    if (AppsPopupH) MemHandleFree((MemHandle) AppsPopupH);

    return newcreator;
}

void CheckPrefsR(void)
{
    /*
     * Do some sanity checks and rewrite the numeric fields to get
     * rid of whitespace or extraneous characters.  Maybe later on
     * I can figure out how to validate this in the input fields.
     */
    numtotxt(StrAToI(PrefsR->POPPrefs.startmsg), tnum, false);
    MemSet(PrefsR->POPPrefs.startmsg, sizeof(PrefsR->POPPrefs.startmsg), 0);
    StrCopy(PrefsR->POPPrefs.startmsg, tnum);
    if (StrAToI(PrefsR->POPPrefs.startmsg) < 1) {
        MemSet(PrefsR->POPPrefs.startmsg,
            sizeof(PrefsR->POPPrefs.startmsg), 0);
        StrCopy(PrefsR->POPPrefs.startmsg,
            DefaultPrefsR.POPPrefs.startmsg);
        PrefsRdirty = true;
    }
    numtotxt(StrAToI(PrefsR->POPPrefs.maxlines), tnum, false);
    MemSet(PrefsR->POPPrefs.maxlines, sizeof(PrefsR->POPPrefs.maxlines), 0);
    StrCopy(PrefsR->POPPrefs.maxlines, tnum);
    if (StrAToI(PrefsR->POPPrefs.maxlines) < 0) {
        MemSet(PrefsR->POPPrefs.maxlines,
            sizeof(PrefsR->POPPrefs.maxlines), 0);
        StrCopy(PrefsR->POPPrefs.maxlines,
            DefaultPrefsR.POPPrefs.maxlines);
        PrefsRdirty = true;
    }
    numtotxt(StrAToI(PrefsR->Prefs.leavefree), tnum, false);
    MemSet(PrefsR->Prefs.leavefree, sizeof(PrefsR->Prefs.leavefree), 0);
    StrCopy(PrefsR->Prefs.leavefree, tnum);
    if (StrAToI(PrefsR->Prefs.leavefree) < 0) {
        MemSet(PrefsR->Prefs.leavefree,
            sizeof(PrefsR->Prefs.leavefree), 0);
        StrCopy(PrefsR->Prefs.leavefree,
            DefaultPrefsR.Prefs.leavefree);
        PrefsRdirty = true;
    }
}

void ReadPrefsRec(void)
{
    Ptr p;
    Int16 i;
    if (MainDB) {
        p = MemHandleLock(PrefsRecHandle);
        MemMove(PrefsR, p, sizeof(struct sPrefsR));
        MemPtrUnlock(p);
        PrefsRdirty = false;
        PrefsWeredirty = false;
        /* Set up the appID array for the icons */
        for (i=0; i < 6; i++) {
            MemMove(&Prefslaunch[i], PrefsR->Prefs.launch[i],
                sizeof(Prefslaunch[i]));
        }
    }
    CheckPrefsR();
}

void WritePrefsRec(void)
{
    Ptr p;
    Err err;
    CheckPrefsR();
    if (MainDB) {
        p = MemHandleLock(PrefsRecHandle);
        err = DmWrite(p, 0, PrefsR, sizeof(struct sPrefsR));
        ErrFatalDisplayIf(err, "Could not write prefs.");
        MemPtrUnlock(p);
        /*
         * This is set until we exit the app at which time the ReleaseRecord()
         * is called.
         */
        PrefsWeredirty = true;
        PrefsRdirty = false;
    }
}

void LoadPrefsFields(void)
{
    FormPtr frm;
	ControlPtr ctl;
	char *label;
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    char name[dmDBNameLength];
    UInt32 AppID;
    Int16 i;

    if ((frm = FrmGetFormPtr(PrefsForm)) == 0) return;
    /* Memory to leave free */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_prefsleavefree)), PrefsR->Prefs.leavefree,
        StrLen(PrefsR->Prefs.leavefree));
    /* Disable auto-off during transfer */
    if (PrefsR->Prefs.noautooff == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsnoautooff)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsnoautooff)), 0);
    /* Attempt text reformatting */
    if (PrefsR->Prefs.reformat == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsreformat)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsreformat)), 0);
    /* Make beep when done */
    if (PrefsR->Prefs.makebeep == '1')
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsmakebeep)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsmakebeep)), 0);
    /* Launches; assumes that IDs are consecutive */
    for (i=0; i < 6; i++) {
        ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            popuptriggerID_prefslaunch1+i));
        label = (char *)CtlGetLabel(ctl);
        MemMove(&AppID, PrefsR->Prefs.launch[i], sizeof(AppID));
        Prefslaunch[i] = AppID;
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, AppID, true, &cardNo, &dbID) == 0) {
            if (DmDatabaseInfo(cardNo, dbID, name, NULL,
                NULL, NULL, NULL,
                NULL, NULL,
                NULL, NULL, NULL, NULL) == 0) {
                StrCopy(label, name);
                CtlSetLabel(ctl, label);
            }
        }
    }
}

void UnloadPrefsFields(Boolean save)
{
    FormPtr frm;
    VoidPtr obj;
    char *ts;
    Int16 i;
    
    if ((frm = FrmGetFormPtr(PrefsForm)) == 0) return;
    /* Memory to leave free */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_prefsleavefree));
    if (save) {
        MemSet(PrefsR->Prefs.leavefree,
            sizeof(PrefsR->Prefs.leavefree), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->Prefs.leavefree, ts, StrLen(ts));
    }
    if (save) {
        /* Disable auto-off during transfer */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsnoautooff))) == 1)
            PrefsR->Prefs.noautooff = '1';
        else PrefsR->Prefs.noautooff = '0';
        /* Attempt text reformatting */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsreformat))) == 1)
            PrefsR->Prefs.reformat = '1';
        else PrefsR->Prefs.reformat = '0';
        /* Make beep when done */
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_prefsmakebeep))) == 1)
            PrefsR->Prefs.makebeep = '1';
        else PrefsR->Prefs.makebeep = '0';
        /* Launches */
        for (i=0; i < 6; i++) {
            MemMove(PrefsR->Prefs.launch[i], &Prefslaunch[i],
                sizeof(Prefslaunch[i]));
        }
        BuildIcons(); /* cause it to rebuild the launch icon records */
    } else {
        /* revert a few settings */
        for (i=0; i < 6; i++) {
            MemMove(&Prefslaunch[i], PrefsR->Prefs.launch[i],
                sizeof(Prefslaunch[i]));
        }
    }
}

void LoadSMTPPrefsFields(void)
{
    FormPtr frm;
    if ((frm = FrmGetFormPtr(SMTPPrefsForm)) == 0) return;
    /* SMTP Server */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpserver)), PrefsR->SMTPPrefs.server,
        StrLen(PrefsR->SMTPPrefs.server));
    /* SMTP Port */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpport)), PrefsR->SMTPPrefs.port,
        StrLen(PrefsR->SMTPPrefs.port));
    /* Your Email */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpemail)), PrefsR->SMTPPrefs.email,
        StrLen(PrefsR->SMTPPrefs.email));
    /* Your Name */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpname)), PrefsR->SMTPPrefs.name,
        StrLen(PrefsR->SMTPPrefs.name));
    /* Additional Bcc */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpaddbcc)), PrefsR->SMTPPrefs.addbcc,
        StrLen(PrefsR->SMTPPrefs.addbcc));
    /* Move to Filed */
    if (PrefsR->SMTPPrefs.flags&SMTPPrefs_FILED)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfiled)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfiled)), 0);
    /* POP auth */
    if (PrefsR->SMTPPrefs.flags&SMTPPrefs_FAKEPOP)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfakepop)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfakepop)), 0);
    /* SMTP auth */
    if (PrefsR->SMTPPrefs.flags&SMTPPrefs_SMTPAUTH)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpauth)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpauth)), 0);
    /* SMTP auth User Name */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpauthuser)), PrefsR->SMTPPrefs.user,
        StrLen(PrefsR->SMTPPrefs.user));
    /* SMTP auth method */
    LstSetSelection(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, 
        listID_smtpauthmethod)), PrefsR->SMTPPrefs.authmethod);
    LstMakeItemVisible( FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, 
        listID_smtpauthmethod)), PrefsR->SMTPPrefs.authmethod);
    CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        popuptriggerID_smtpauthmethod)),
        LstGetSelectionText(FrmGetObjectPtr(frm,
        FrmGetObjectIndex(frm, 
        listID_smtpauthmethod)), PrefsR->SMTPPrefs.authmethod));
}

void UnloadSMTPPrefsFields(Boolean save)
{
    FormPtr frm;
    VoidPtr obj;
    char *ts;
    if ((frm = FrmGetFormPtr(SMTPPrefsForm)) == 0) return;
    /* SMTP Server */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpserver));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.server,
            sizeof(PrefsR->SMTPPrefs.server), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.server, ts, StrLen(ts));
    }
    /* SMTP Port */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpport));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.port,
            sizeof(PrefsR->SMTPPrefs.port), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.port, ts, StrLen(ts));
    }
    /* Your Email */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpemail));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.email,
            sizeof(PrefsR->SMTPPrefs.email), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.email, ts, StrLen(ts));
    }
    /* Your Name */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpname));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.name,
            sizeof(PrefsR->SMTPPrefs.name), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.name, ts, StrLen(ts));
    }
    /* Additional Bcc */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpaddbcc));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.addbcc,
            sizeof(PrefsR->SMTPPrefs.addbcc), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.addbcc, ts, StrLen(ts));
    }
    /* Move to Filed */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfiled))) == 1)
            PrefsR->SMTPPrefs.flags |= SMTPPrefs_FILED;
        else PrefsR->SMTPPrefs.flags &= ~SMTPPrefs_FILED;
    }
    /* POP auth */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_smtpfakepop))) == 1)
            PrefsR->SMTPPrefs.flags |= SMTPPrefs_FAKEPOP;
        else PrefsR->SMTPPrefs.flags &= ~SMTPPrefs_FAKEPOP;
    }
    /* SMTP auth */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
	    checkboxID_smtpauth))) == 1)
            PrefsR->SMTPPrefs.flags |= SMTPPrefs_SMTPAUTH;
        else PrefsR->SMTPPrefs.flags &= ~SMTPPrefs_SMTPAUTH;
    }
    /* SMTP auth User Name */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_smtpauthuser));
    if (save) {
        MemSet(PrefsR->SMTPPrefs.user,
            sizeof(PrefsR->SMTPPrefs.user), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->SMTPPrefs.user, ts, StrLen(ts));
    }
    /* SMTP auth method */
    if (save) {
        PrefsR->SMTPPrefs.authmethod = LstGetSelection(FrmGetObjectPtr(frm,
            FrmGetObjectIndex(frm, listID_smtpauthmethod)));
    }
}

void LoadPOPPrefsFields(void)
{
    FormPtr frm;
    if ((frm = FrmGetFormPtr(POPPrefsForm)) == 0) return;
    /* POP Server */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popserver)), PrefsR->POPPrefs.server,
        StrLen(PrefsR->POPPrefs.server));
    /* POP Port */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popport)), PrefsR->POPPrefs.port,
        StrLen(PrefsR->POPPrefs.port));
    /* User */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popuser)), PrefsR->POPPrefs.user,
        StrLen(PrefsR->POPPrefs.user));
    /* Max msg lines */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popmaxlines)), PrefsR->POPPrefs.maxlines,
        StrLen(PrefsR->POPPrefs.maxlines));
    /* Starting message number */
    FldInsert(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popstartmsg)), PrefsR->POPPrefs.startmsg,
        StrLen(PrefsR->POPPrefs.startmsg));
    /* Detect mailbox reset */
    if (PrefsR->POPPrefs.flags&POPPrefs_DETECTRESET)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdetectreset)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdetectreset)), 0);
    /* Delete messages from server */
    if (PrefsR->POPPrefs.flags&POPPrefs_DELETEMSG)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdeletemsg)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdeletemsg)), 0);
    /* Use APOP for authentication */
    if (PrefsR->POPPrefs.flags&POPPrefs_USEAPOP)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popuseapop)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popuseapop)), 0);
    /* Scheduled POP */
    if (PrefsR->POPPrefs.flags&POPPrefs_SCHEDULEDPOP)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popscheduledpop)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popscheduledpop)), 0);
    /* Use HTTP */
    if (PrefsR->POPPrefs.flags&POPPrefs_USEHTTP)
        CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popusehttp)), 1);
    else CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popusehttp)), 0);
}

void UnloadPOPPrefsFields(Boolean save)
{
    FormPtr frm;
    VoidPtr obj;
    char *ts;
    if ((frm = FrmGetFormPtr(POPPrefsForm)) == 0) return;
    /* POP Server */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popserver));
    if (save) {
        MemSet(PrefsR->POPPrefs.server,
            sizeof(PrefsR->POPPrefs.server), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->POPPrefs.server, ts, StrLen(ts));
    }
    /* POP Port */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popport));
    if (save) {
        MemSet(PrefsR->POPPrefs.port,
            sizeof(PrefsR->POPPrefs.port), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->POPPrefs.port, ts, StrLen(ts));
    }
    /* User */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popuser));
    if (save) {
        MemSet(PrefsR->POPPrefs.user,
            sizeof(PrefsR->POPPrefs.user), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->POPPrefs.user, ts, StrLen(ts));
    }
    /* Max msg lines */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popmaxlines));
    if (save) {
        MemSet(PrefsR->POPPrefs.maxlines,
            sizeof(PrefsR->POPPrefs.maxlines), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->POPPrefs.maxlines, ts, StrLen(ts));
    }
    /* Starting message number */
    obj = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
        fieldID_popstartmsg));
    if (save) {
        MemSet(PrefsR->POPPrefs.startmsg,
            sizeof(PrefsR->POPPrefs.startmsg), 0);
        if ((ts = FldGetTextPtr(obj)))
            MemMove(PrefsR->POPPrefs.startmsg, ts, StrLen(ts));
    }
    /* Detect mailbox reset */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdetectreset))) == 1)
            PrefsR->POPPrefs.flags |= POPPrefs_DETECTRESET;
        else PrefsR->POPPrefs.flags &= ~POPPrefs_DETECTRESET;
    }
    /* Delete messages from server */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popdeletemsg))) == 1)
            PrefsR->POPPrefs.flags |= POPPrefs_DELETEMSG;
        else PrefsR->POPPrefs.flags &= ~POPPrefs_DELETEMSG;
    }
    /* Use APOP for authentication */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popuseapop))) == 1)
            PrefsR->POPPrefs.flags |= POPPrefs_USEAPOP;
        else PrefsR->POPPrefs.flags &= ~POPPrefs_USEAPOP;
    }
    /* Scheduled POP */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popscheduledpop))) == 1) {
            PrefsR->POPPrefs.flags |= POPPrefs_SCHEDULEDPOP;
            register_sleep();
            register_wake();
        } else {
            PrefsR->POPPrefs.flags &= ~POPPrefs_SCHEDULEDPOP;
            unregister_sleep();
            register_wake();
        }
    }
    /* Use HTTP */
    if (save) {
        if (CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
            checkboxID_popusehttp))) == 1)
            PrefsR->POPPrefs.flags |= POPPrefs_USEHTTP;
        else PrefsR->POPPrefs.flags &= ~POPPrefs_USEHTTP;
    }
}

Boolean is_os_32(void)
{
    UInt32 romVersion;
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    /* some NotifyMgr things need >= 3.2 */
    if (romVersion >= 0x03200000) return true;
    return false;
}

void register_wake(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyRegister(cardNo, dbID, sysNotifyLateWakeupEvent, NULL,
        sysNotifyNormalPriority, NULL);
}

void unregister_wake(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 almtime;
    UInt32 ref;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyUnregister(cardNo, dbID, sysNotifyLateWakeupEvent,
        sysNotifyNormalPriority);
    if ((almtime = GetTimeOfNextAlarm(&ref, NULL)) == 0) return;
    /* cancel our alarm if it wasn't what awakened us */
    if (almtime > TimGetSeconds())
        SetTimeOfNextAlarm(0, 0, NULL);
}

void register_sleep(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyRegister(cardNo, dbID, sysNotifySleepNotifyEvent, NULL,
        sysNotifyNormalPriority, NULL);
}

void unregister_sleep(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (!is_os_32()) return;
    if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
        sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
    SysNotifyUnregister(cardNo, dbID, sysNotifySleepNotifyEvent,
        sysNotifyNormalPriority);
}

DmOpenRef mailapp_get_db(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, MailAppID, true, &cardNo, &dbID) == 0)) {
        SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdCustomBase+1,
            NULL, &res);
        return (DmOpenRef)res; /* non-zero if Mail app is running */
    }
    return 0;
}

void mailapp_refresh_list(void)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, MailAppID, true, &cardNo, &dbID) == 0)) {
        SysAppLaunch(cardNo, dbID, 0, sysAppLaunchCmdCustomBase+2,
            NULL, &res);
    }
}

void post_custom_event(UInt16 etype)
{
    EventType new_event;
    new_event.eType = etype;
    EvtAddEventToQueue(&new_event);
}

/*
 * The program starts here.
 */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    Err err;
    UInt32 romVersion;
    UInt32 theInstalledRAM; 
    FormPtr frm;
    FontID fnt;
    UInt32 ref;
    UInt32 almtime, curtime, lasttime;
    struct sPrefsR *prefsr;
    DmOpenRef maindb;
    Int16 prefslocked;
    UInt32 smstrig;
    UInt16 tui16;

    switch (cmd) {
    case sysAppLaunchCmdNormalLaunch:
        /*
         * This ROM version test comes from Palm's Address application.
         * Thankfully too because I would hate to guess at this kind of
         * stuff and make some poor guy's Pilot blow up.
         */
        FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
        /*
         * Need at least a PalmPilot Pro because of the TCP/IP.
         */
        if (romVersion < 0x03000000) {
            if ((launchFlags &
                (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
                (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
                /*
                 * Pilot 1.0 will continuously relaunch this app unless we
                 * switch to another safe one.
                 */
                if (romVersion < 0x02000000) {
                    FrmAlert(RomIncompatibleAlert);
                    AppLaunchWithCommand(sysFileCDefaultApp,
                        sysAppLaunchCmdNormalLaunch, NULL);
                }
                err = MemCardInfo( 0, NULL, NULL, NULL, NULL, NULL,
                    &theInstalledRAM, NULL); 
                if (!err && (theInstalledRAM == 524288)) { 
                    /*
                     * We are running on a PalmPilot Personal unit
                     */
                    FrmAlert(RomIncompatibleAlert);
                    return sysErrRomIncompatible;
                }
            }
        }
        /*
         * I guess we are an appropriate ROM version so let's run.
         */
        err = StartApplication();
        if (err) return err;
        /* register for sleep/wake notifications */
        if (PrefsR->POPPrefs.flags&POPPrefs_SCHEDULEDPOP) {
            register_sleep();
            register_wake();
        }
        EventLoop();
        StopApplication();
        break;
    case sysAppLaunchCmdAlarmTriggered:
        AlarmTriggered((SysAlarmTriggeredParamType *)cmdPBP);
        break;
    case sysAppLaunchCmdDisplayAlarm:
        DisplayAlarm(launchFlags);
        break;
    case sysAppLaunchCmdCustomBase+1: /* background download */
        if (!cmdPBP) {
            /* this is from my DisplayAlarm() */
            tui16 = EV_CHECKMAIL;
        } else {
            /* this parameter tells me whether it is a Check or Send */
            switch (*(UInt8 *)cmdPBP) {
            case 0: /* Check Mail */
                tui16 = EV_CHECKMAIL;
                /*
                 * Manual mail check should cancel out any SMS trigger.  Just
                 * for convenience so it doesn't check twice.
                 */
                if (!FtrGet(MainAppID, FTR_SMSTRIG, &smstrig)) {
                    /* no more alarms from this SMS */
                    FtrUnregister(MainAppID, FTR_SMSTRIG);
                    unregister_sleep();
                    unregister_wake();
                }
                break;
            case 1: tui16 = EV_SENDMAIL; break; /* Send Mail */
            default: tui16 = EV_CHECKMAIL; break;
            }
        }

        /* currently in application? */
        if (!(launchFlags&sysAppLaunchFlagNewGlobals)) {
            /* so post an event to start POP */
            post_custom_event(tui16);
            return 0;
        }
  
        /*
         * Now I do something sneaky.  I want to be able to do a download
         * while the Mail application is running.  So I've added a couple
         * launch codes to it to allow me to share the database, and
         * indicate that Mail must refresh it's list.
         */
        if ((MailDB = mailapp_get_db())) ismymaildb = false;

        isbackground = true;
        backgroundretval = RV_SUCCESS;
        err = StartApplication();
        if (err) return err;
        fnt = FntSetFont(stdFont);
        frm = FrmInitForm(BackgroundForm);
        FrmSetActiveForm(frm);
        FrmSetEventHandler(frm, BackgroundFormHandleEvent);
        FrmDrawForm(frm);
        ShowStatusLine(true);
        switch (tui16) {
        case EV_CHECKMAIL:
            if ((backgroundretval = OpenPOP3(0))) {
                post_custom_event(EV_STOPLOOP);
            }
            break;
        case EV_SENDMAIL:
            if (!OpenSMTP()) {
                post_custom_event(EV_STOPLOOP);
                backgroundretval = RV_FAILED;
            }
            break;
        }
        EventLoop();
        StopApplication();
        FrmEraseForm(frm);
        FrmDeleteForm(frm);
        FntSetFont(fnt);
        FrmSetActiveForm(NULL);
        return backgroundretval;
    case sysAppLaunchCmdCustomBase+2: /* initiate download from another app */
        /*
         * I don't want to do anything while the device is on.  So I set a
         * feature which is checked when the device is turned off.  Basically
         * the same as is done with the scheduled POP, except the interval
         * is different.
         */
        FtrSet(MainAppID, FTR_SMSTRIG, 5 /* seconds */);
        register_sleep();
        register_wake();
        break;
    case sysAppLaunchCmdNotify:
        /*
         * Turn off scheduled POP when awake.  Turn it back on
         * when asleep.
         */
        switch (((SysNotifyParamType*)cmdPBP)->notifyType) {
        case sysNotifySleepNotifyEvent:
            /* we want to prevent alarm stacking */
            if (!FtrGet(MainAppID, FTR_POPACTIVE, &smstrig)) break;
            /* set an alarm for the next interval */
            curtime = TimGetSeconds();
            /*
             * First off I'm going to determine if an interval has
             * passed while the user was using the device.  I don't
             * check mail while the device is on, so if the user
             * happens to be using the device when an interval passes,
             * he will have to wait a full interval for the next check.
             *
             * This is an attempt at improving this situation, by shortening
             * the next interval if one was skipped due to the user.
             */
            do {
            if (!FtrGet(MainAppID, FTR_LASTTIME, &lasttime)) {
                /* FTR_LASTTIME was stored in our last successful PollForMail */
                if (curtime-lasttime > (UInt32)BACKGROUNDINTERVAL) {
                    /*
                     * An interval has passed due to user activity.  So let's
                     * set the next interval to just two minutes.
                     */
                    //almtime = (UInt32)(2*60)-(curtime%((UInt32)(2*60)));
                    //break;
                }
            }
            /* otherwise set the interval as expected */
            almtime = (UInt32)BACKGROUNDINTERVAL-
                (curtime%((UInt32)BACKGROUNDINTERVAL));
            } while (0);
            /* we may be here because of sysAppLaunchCmdCustomBase+2 */
            if (!FtrGet(MainAppID, FTR_SMSTRIG, &smstrig)) {
                /* so I'll set it to start a POP in xxx secs */
                almtime = smstrig;
            }
            SetTimeOfNextAlarm(curtime+almtime, 0, NULL);
            break;
        case sysNotifyLateWakeupEvent:
            almtime = GetTimeOfNextAlarm(&ref, NULL);
            /* cancel our alarm if it wasn't what awakened us */
            if (!almtime || (almtime > TimGetSeconds())) {
                /* comes here during any wakeup not caused by us */
                if (almtime) SetTimeOfNextAlarm(0, 0, NULL);
                break;
            }
            /* comes here during wakeup caused by us */
        }
        break;
    case sysAppLaunchCmdSystemReset:
    case sysAppLaunchCmdSyncNotify:
    case sysAppLaunchCmdTimeChange:
        /* register for sleep/wake notifications */
        if ((prefslocked = lock_prefs_without_globals(&maindb, &prefsr))) {
            if (prefsr->POPPrefs.flags&POPPrefs_SCHEDULEDPOP) {
                register_sleep();
                register_wake();
            }
            if (prefslocked == 1) unlock_prefs_without_globals(maindb, prefsr);
        }
        break;
    default:
        return sysErrParamErr;
    }

    return 0;
}
