/* $Id: gnugotmail.h,v 1.95 2004/03/12 20:48:00 chrisf Exp $ */

/*
GNU Got Mail - An SMTP/POP3 client for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __GNUGOTMAIL_H__
#define __GNUGOTMAIL_H__

#include "sdkver.h"
#include "gnugotmailDB.h"

/*
 * This flag (0 or 1) controls the ShowDebugMsg().. basically just showing
 * some extra messages on the screen.  Not extremely useful.
 */
#define DEBUGMSG 0

#define MainAppID 'GNGM'
#define MainDBType 'DATA'

/*
 * These are X-Mailer and X-Url headers that are pasted for outgoing mail.
 */
#define XMAILER "GNU Got Mail 0.83 for PalmPilot"
#define XURL "http://rallypilot.sourceforge.net"

#define AppErrStrLen 32
#define ErrToString(err) SysErrString(err,AppErrStr,AppErrStrLen)

/*
 * MSGRECSIZE is the amount of memory to expand an Inbox record at one time.
 * This used to be small (2K) but I found that it would sneak into small
 * spaces and cause out of memory problems.  Also it must be large enough to
 * fit all of the header fields at one time (prob at least 1.5K).
 */
#define MSGRECSIZE 8192

/*
 * This needs to be big enough to hold the largest amount of chars that are
 * to be send at one time.  The sendq is completely flushed before the tick
 * functions (poptick(), smtptick()) are called each time.. so that they are
 * ensured to be able to fit their outgoing chars without checking.
 */
#define SENDQSIZE 256

/* How many seconds between (from finish time) background POPs */
#define BACKGROUNDINTERVAL 30*60

/* These are temporary features */
#define FTR_HOSTHASH 1
#define FTR_HOSTIP 2
#define FTR_LASTTIME 3
#define FTR_SMSTRIG 4
#define FTR_POPACTIVE 5

/* Return value types for our scheduled mailcheck */
#define RV_SUCCESS 0
#define RV_FAILED 1
#define RV_USERCANCEL 2

/* Custom events */
#define EV_STOPLOOP (firstUserEvent+1)
#define EV_CHECKMAIL (firstUserEvent+2)
#define EV_SENDMAIL (firstUserEvent+3)

/* function prototypes */

void MigratePrefs(struct sPrefsR *currP, struct s12PrefsR *oldP);
Boolean MainFormHandleEvent(EventPtr e);
Boolean PrefsFormHandleEvent(EventPtr e);
Boolean SMTPPrefsFormHandleEvent(EventPtr e);
Boolean SMTPPrefsPassFormHandleEvent(EventPtr e);
Boolean POPPrefsFormHandleEvent(EventPtr e);
Boolean POPPrefsPassFormHandleEvent(EventPtr e);
Boolean AboutFormHandleEvent(EventPtr e);
Boolean ContribFormHandleEvent(EventPtr e);
Boolean BackgroundFormHandleEvent(EventPtr e);
FieldPtr GetFocusObjectPtr(void);
Boolean FormHandleMenuEvent(UInt16 menuID);
Boolean ApplicationHandleEvent(EventPtr e);
void freememories(void);
UInt16 StartApplication(void);
void StopApplication(void);
void EventLoop(void);
UInt32 WhichPhoneType(void);
Err LoadPhoneLibrary(UInt16 *plref);
void ClosePhoneLibrary(UInt16 plref);
Boolean is_ok_to_connect(void);
Err OpenPOP3(Int16 whattodo);
void ClosePOP3(void);
Boolean OpenSMTP(void);
void CloseSMTP(void);
Boolean OpenDatabases(void);
void CloseDatabases(void);
void ClearInbox(void);
void AppsPopupDrawItem(UInt16 itemNum, RectanglePtr bounds,
    char **unusedP);
UInt32 AppsPopup(UInt16 listid, UInt16 triggerid, UInt32 curcreator);
void CheckPrefsR(void);
void ReadPrefsRec(void);
void WritePrefsRec(void);
void LoadPrefsFields(void);
void UnloadPrefsFields(Boolean save);
void LoadSMTPPrefsFields(void);
void UnloadSMTPPrefsFields(Boolean save);
void LoadPOPPrefsFields(void);
void UnloadPOPPrefsFields(Boolean save);
Boolean is_os_32(void);
void register_wake(void);
void unregister_wake(void);
void register_sleep(void);
void unregister_sleep(void);
DmOpenRef mailapp_get_db(void);
void mailapp_refresh_list(void);
void post_custom_event(UInt16 etype);
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags);

#endif
