/* $Id: gnugotmailDB.h,v 1.25 2003/10/04 07:24:52 chrisf Exp $ */

/*
 * This is a structure of a MailDB record.
 *
 * DateType date;
 * TimeType time;
 * MailDBRecordFlags flags;
 * char subject[], from[], to[], cc[], bcc[], replyTo[], sentTo[], body[];
 *
 * And here is a structure with just the date, time, and flags so that I
 * can tell where the message data starts.
 */

#ifndef __GNUGOTMAILDB_H__
#define __GNUGOTMAILDB_H__

#define PREFSVERSION 13

#define MailAppID 'mail'
#define MailDBType 'DATA'
#define MailSigPrefID 3

typedef struct {
        UInt16    read            : 1;
        UInt16    signature       : 1;
        UInt16    confirmRead     : 1;
        UInt16    confirmDelivery : 1;
        UInt16    priority        : 2; /* MailPriorityType */
        UInt16    addressing      : 2; /* MailAddressingType */
        UInt16    reserved        : 8;
} MailFlagsType;

typedef enum { priorityHigh, priorityNormal, priorityLow } MailPriorityType;
typedef enum { sentTo, sentCC, sentBCC } MailAddressingType;

struct sMailDB {
    DateType date;
    TimeType time;
    MailFlagsType flags;
    char *msg;
};

#define sortByDateItem      0
#define sortByFromItem      1
#define sortBySubjectItem   2

/* General Preferences */
struct sPrefs {
    char leavefree[6];
    char noautooff;
    char reformat;
    char makebeep;
    char launch[6][4];
};

/* The SMTP preferences */
struct sSMTPPrefs {
    char server[65];
    char port[6];
    char email[65];
    char name[65];
    char addbcc[65];
    #define SMTPPrefs_FILED (1<<0)
    #define SMTPPrefs_FAKEPOP (1<<1)
    #define SMTPPrefs_SMTPAUTH (1<<2)
    char flags;
    char authmethod;
    char user[65];
    char passwd[65];
};

/* One of the POP preferences */
struct sPOPPrefs {
    char server[65];
    char port[6];
    char user[65];
    char pass[17];
    char maxlines[9];
    char startmsg[9];
    #define POPPrefs_DETECTRESET (1<<0)
    #define POPPrefs_DELETEMSG (1<<1)
    #define POPPrefs_USEAPOP (1<<2)
    #define POPPrefs_SCHEDULEDPOP (1<<3)
    #define POPPrefs_USEHTTP (1<<4)
    #define POPPrefs_HTTPSUCCESS (1<<5)
    char flags;
    char firstuidl[81];
};

/*
 * A preferences record which consists of back-to-back
 * structures.  It also has a version char which starts at 1 to
 * allow testing for preferences migration in future versions.
 */
struct sPrefsR {
    char version;
    struct sPrefs Prefs;
    struct sSMTPPrefs SMTPPrefs;
    struct sPOPPrefs POPPrefs;
};

/*
 * THESE ARE THE STRUCTURES FOR THE PREVIOUS VERSION'S PREFERENCES.  I USE
 * THEM FOR MIGRATION PURPOSES.
 */

/* General Preferences */
struct s12Prefs {
    char leavefree[6];
    char noautooff;
    char reformat;
    char makebeep;
    char launch[6][4];
};

/* The SMTP preferences */
struct s12SMTPPrefs {
    char server[65];
    char port[6];
    char email[65];
    char name[65];
    char addbcc[65];
    char filed;
    char fakepop;
};

/* One of the POP preferences */
struct s12POPPrefs {
    char server[65];
    char port[6];
    char user[65];
    char pass[17];
    char maxlines[9];
    char startmsg[9];
    char detectreset;
    char deletemsg;
    char useapop;
    char firstuidl[81];
};

struct s12PrefsR {
    char version;
    struct s12Prefs Prefs;
    struct s12SMTPPrefs SMTPPrefs;
    struct s12POPPrefs POPPrefs;
};

/*******************************************************/

#endif
