/* $Id: sockets.c,v 1.35 2003/10/05 01:43:58 chrisf Exp $ */

/*
GNU Got Mail - An SMTP/POP3 client for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <PalmOS.h>
#include <System/NetMgr.h>
#include "common.h"
#include "gnugotmail.h"
#include "sockets.h"
#include "pop.h"

/* define some data types */

extern Int32 AppNetTimeout;
extern UInt16 AppNetRefnum;
extern Err errno;
extern Char AppErrStr[];
extern struct spopdata popd;
extern struct ssmtpdata smtpd;

/* globals */

char localip[20], remoteip[20];
char base64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

UInt32 rotating_hash(char *key, Int16 len)
{
    UInt32 hash;
    Int16 i;
    for (hash=len, i=0; i < len; ++i) hash = (hash<<4)^(hash>>28)^key[i];
    return hash;
}

Int16 connectsock(char *host, char *service, Boolean nonblockconnect)
{
    NetSocketRef sock;
    Int16 one;
    Int16 i1;
    NetSocketAddrINType saddr;
    NetHostInfoPtr phe;
    NetHostInfoBufType AppHostInfo;
    char tempbuf[64];
    Boolean nameresolved=false;
    UInt32 hosthash;
    Int16 ret;

    one = 1;

    MemSet((char *) &saddr, sizeof(saddr), 0);
    saddr.family = netSocketAddrINET;
    saddr.port = NetHToNS(StrAToI(service));

    if ((sock = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET,
        netSocketTypeStream, 0, AppNetTimeout, &errno)) < 0) {
        ShowInfoMsg("");
        SetErrorMsg(ErrToString(errno));
        return 0;
    }

    /* see if we can use our DNS cache */
    if (!FtrGet(MainAppID, FTR_HOSTHASH, &hosthash) &&
        (rotating_hash(host, StrLen(host)) == hosthash) &&
        !FtrGet(MainAppID, FTR_HOSTIP, (UInt32 *)&saddr.addr)) {
        nameresolved = true;
    }

    /*
     * Map host name to IP address, allowing for dotted decimal.
     * I try to assimilate the name first to see if it is just
     * a dotted IP.  If not, then I use the DNS to resolve.
     *
     * Also it tries to retrieve a cached IP from a feature.
     */
    ShowInfoMsg("Looking up IP.");
    if (!nameresolved ||
        ((saddr.addr = NetLibAddrAToIN(AppNetRefnum, host)) == -1)) {
        phe = NetLibGetHostByName(AppNetRefnum, host, &AppHostInfo,
            AppNetTimeout, &errno);
        if (!errno) {
            /*
             * Sometimes the first address in the list is zero.  This
             * makes me have to search through them to find the first
             * non-zero address.
             */
            for (i1=0; i1 < netDNSMaxAddresses; i1++) {
                if (phe->addrListP[i1] != NULL) {
                    if (*phe->addrListP[i1] != '\0') {
                        MemMove((char *) &saddr.addr,
                            (char *) phe->addrListP[i1], phe->addrLen);
                        nameresolved = true;
                        break;
                    }
                }
            }
        }
        if (!nameresolved) {
            NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &errno);
            ShowInfoMsg("");
            SetErrorMsg("NetLibGetHostByName failure.");
            return 0;
        } else {
            /*
             * Cache the IP to make future lookups quicker.  Note that
             * I store a hash of the hostname only to validate the need for
             * an uncached DNS lookup later.
             */
            FtrSet(MainAppID, FTR_HOSTHASH, rotating_hash(host, StrLen(host)));
            FtrSet(MainAppID, FTR_HOSTIP, (UInt32)saddr.addr);
        }
    }

    /* set to non-blocking before connect */
    if (nonblockconnect) {
        NetLibSocketOptionSet(AppNetRefnum, sock, netSocketOptLevelSocket,
            netSocketOptSockNonBlocking, &one, sizeof(one), AppNetTimeout,
            &errno);
    }

    /* initiate a connect */
    ret = NetLibSocketConnect(AppNetRefnum, sock, (NetSocketAddrType *) &saddr,
        sizeof(saddr), AppNetTimeout, &errno);
    if (ret && (errno != netErrWouldBlock)) {
        NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &errno);
        ShowInfoMsg("");
        SetErrorMsg("connect failure.");
        return 0;
    }

    /* set to non-blocking after connect */
    if (!nonblockconnect) {
        NetLibSocketOptionSet(AppNetRefnum, sock, netSocketOptLevelSocket,
            netSocketOptSockNonBlocking, &one, sizeof(one), AppNetTimeout,
            &errno);
    }

    /* turn off Nagle */
    NetLibSocketOptionSet(AppNetRefnum, sock, netSocketOptLevelTCP,
        netSocketOptTCPNoDelay, &one, sizeof(one), AppNetTimeout, &errno);

    /*
     * Get the local and remote IP addresses.  This is used for the HELO
     * SMTP command later.  Actually I got rid of these because I wanted
     * to go to a non-blocking connect.  NetLibSocketAddr() would have to
     * be moved to other places because it requires a connected socket.
     */

    StrCopy(localip, "foo.bar");
    remoteip[0] = (char) NULL;

    StrCopy(tempbuf, "Connect: ");
    StrCat(tempbuf, host);
    ShowInfoMsg(tempbuf);

    return sock;
} 

/* I pass ref & timeout because I use this without globals */
Int16 connectsockudp(char *host, char *service, UInt16 refnum, Int32 timeout)
{
    NetSocketRef sock;
    Int16 one;
    Int16 i1;
    NetSocketAddrINType saddr;
    NetHostInfoPtr phe;
    NetHostInfoBufType AppHostInfo;
    Boolean nameresolved=false;
    UInt32 hosthash;

    one = 1;

    MemSet((char *) &saddr, sizeof(saddr), 0);
    saddr.family = netSocketAddrINET;
    saddr.port = NetHToNS(StrAToI(service));

    if ((sock = NetLibSocketOpen(refnum, netSocketAddrINET,
        netSocketTypeDatagram, 0, timeout, &errno)) < 0) {
        return 0;
    }

    /* see if we can use our DNS cache */
    if (!FtrGet(MainAppID, FTR_HOSTHASH, &hosthash) &&
        (rotating_hash(host, StrLen(host)) == hosthash) &&
        !FtrGet(MainAppID, FTR_HOSTIP, (UInt32 *)&saddr.addr)) {
        nameresolved = true;
    }

    /*
     * Map host name to IP address, allowing for dotted decimal.
     * I try to assimilate the name first to see if it is just
     * a dotted IP.  If not, then I use the DNS to resolve.
     *
     * Also it tries to retrieve a cached IP from a feature.
     */
    if (!nameresolved ||
        ((saddr.addr = NetLibAddrAToIN(refnum, host)) == -1)) {
        if ((phe = NetLibGetHostByName(refnum, host, &AppHostInfo,
            timeout, &errno)) != 0) {
            /*
             * Sometimes the first address in the list is zero.  This
             * makes me have to search through them to find the first
             * non-zero address.
             */
            for (i1=0; i1 < netDNSMaxAddresses; i1++) {
                if (phe->addrListP[i1] != NULL) {
                    if (*phe->addrListP[i1] != '\0') {
                        MemMove((char *) &saddr.addr,
                            (char *) phe->addrListP[i1], phe->addrLen);
                        nameresolved = true;
                        break;
                    }
                }
            }
        }
        if (!nameresolved) {
            NetLibSocketClose(refnum, sock, timeout, &errno);
            return 0;
        } else {
            /*
             * Cache the IP to make future lookups quicker.  Note that
             * I store a hash of the hostname only to validate the need for
             * an uncached DNS lookup later.
             */
            FtrSet(MainAppID, FTR_HOSTHASH, rotating_hash(host, StrLen(host)));
            FtrSet(MainAppID, FTR_HOSTIP, (UInt32)saddr.addr);
        }
    }

    if (NetLibSocketConnect(refnum, sock, (NetSocketAddrType *) &saddr,
        sizeof(saddr), timeout, &errno) != 0) {
        NetLibSocketClose(refnum, sock, timeout, &errno);
        return 0;
    }

    /* set to non-blocking */
    NetLibSocketOptionSet(refnum, sock, netSocketOptLevelSocket,
        netSocketOptSockNonBlocking, &one, sizeof(one), timeout, &errno);
    /* turn off Nagle */
    NetLibSocketOptionSet(refnum, sock, netSocketOptLevelTCP,
        netSocketOptTCPNoDelay, &one, sizeof(one), timeout, &errno);

    /*
     * This usually fails in the emulator.  It's needed to have PalmOS
     * assign a non-zero source port.
     */
    MemSet((char *) &saddr, sizeof(saddr), 0);
    NetLibSocketBind(refnum, sock, (NetSocketAddrType *) &saddr,
        sizeof(saddr), timeout, &errno);

    return sock;
} 

Int16 writeall(Int16 sock, struct ssq *sq, char *str, Int16 len)
{
    Int16 i1;
    i1 = MIN(SENDQSIZE-sq->sendqin, len);
    if (i1 <= 0) return 0;
    MemMove(&sq->sendq[sq->sendqin], str, i1);
    sq->sendqin += i1;
    return i1;
}

/*
 * This encodes with base-64 as it sticks data into the queue.  Make sure
 * you call it with 3 byte multiples, unless you want it padded -- like at
 * the end of a transmission.  Also this must be guaranteed to fit in the
 * sendq.. so you must've known to check if 4/3 * len will fit.
 */
Int16 writeallb64(Int16 sock, struct ssq *sq, char *str, Int16 len)
{
    Int16 i1, i2;
    UInt32 l1, l2, l3;
    i1 = MIN(SENDQSIZE-sq->sendqin, len);
    if (i1 <= 0) return 0;
    for (i2=0; i2 <= (len-3); i2=i2+3) {
        l1 = str[i2] & 0xff;
        l2 = str[i2+1] & 0xff;
        l3 = str[i2+2] & 0xff;
        l1 = (l1 << 16)|(l2 << 8)|l3;
        sq->sendq[sq->sendqin++] = base64[(l1 >> 18) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[(l1 >> 12) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[(l1 >> 6) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[l1 & 0x3f];
    }
    /* pad the remainder with = */
    switch(len % 3) {
    case 2:
        l1 = str[len-2] & 0xff;
        l2 = str[len-1] & 0xff;
        l1 = (l1 << 8)|l2;
        sq->sendq[sq->sendqin++] = base64[(l1 >> 10) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[(l1 >> 4) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[(l1 << 2) & 0x3c];
        sq->sendq[sq->sendqin++] = '=';
        break;
    case 1:
        l1 = str[len-1] & 0xff;
        sq->sendq[sq->sendqin++] = base64[(l1 >> 2) & 0x3f];
        sq->sendq[sq->sendqin++] = base64[(l1 << 4) & 0x30];
        sq->sendq[sq->sendqin++] = '=';
        sq->sendq[sq->sendqin++] = '=';
    }
    return len;
}
