/* $Id: timer.c,v 1.73 2003/11/23 04:40:57 chrisf Exp $ */

/*
GNU Got Mail - An SMTP/POP3 client for the PalmPilot
Copyright (C) 1999 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <PalmOS.h>
#include "callback.h"
#include "gnugotmail.h"
#include "gnugotmailRsc.h"
#include "gnugotmailDB.h"
#include "common.h"
#include "sockets.h"
#include "timer.h"

extern DmOpenRef MainDB;
extern struct sPrefsR *PrefsR;

void SetTimeOfNextAlarm(UInt32 alarmTime, UInt32 ref, AlmAlarmProcPtr procP)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (procP) {
        AlmSetProcAlarm(procP, ref, alarmTime);
    } else {
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return;
        AlmSetAlarm(cardNo, dbID, ref, alarmTime, 0);
    }
}

UInt32 GetTimeOfNextAlarm(UInt32 *refP, AlmAlarmProcPtr procP)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;

    if (procP) {
        return AlmGetProcAlarm(procP, refP);
    } else {
        if (DmGetNextDatabaseByTypeCreator(true, &searchInfo, 
            sysFileTApplication, MainAppID, true, &cardNo, &dbID)) return 0;
        return AlmGetAlarm(cardNo, dbID, refP);
    }
}

Int16 lock_prefs_without_globals(DmOpenRef *maindb, struct sPrefsR **prefsr)
{
    MemHandle prefsh;
    void *p;
//    SysAppInfoPtr sai;

    /* can't get SysGetAppInfo to work on my Treo 300 */
    /* first we need to determine whether globals are available */
//    if (!(sai = SysGetAppInfo())) return 0; /* failure */
//    if (sai->launchFlags&sysAppLaunchFlagSubCall) {
//        /* we're currently in our app, so globals are available */
//        *maindb = MainDB;
//        *prefsr = PrefsR;
//        return 2; /* this indicates not to try and close/unlock them */
//    }

    /* we're not in our app, so open the db ourselves and lock the prefs */
    if ((*maindb = DmOpenDatabaseByTypeCreator(MainDBType, MainAppID,
        dmModeReadWrite))) {
        if ((prefsh = DmQueryRecord(*maindb, 0)) &&
            (p = MemHandleLock(prefsh)) &&
            (MemHandleSize(prefsh) >= sizeof(char)) &&
            (*(char *)p == (char)PREFSVERSION)) {
            *prefsr = p;
        } else {
            DmCloseDatabase(*maindb);
            return 0; /* failure */
        }
    } else return 0; /* failure */
    return 1; /* this indicates that we should close/unlock when done */
}

void unlock_prefs_without_globals(DmOpenRef maindb, struct sPrefsR *prefsr)
{
    /* unlock our prefs record */
    MemPtrUnlock(prefsr);
    DmCloseDatabase(maindb);
}

void PollForMail(UInt16 almProcCmd, SysAlarmTriggeredParamType *paramP)
{
    Err err, ifErrs;
    UInt16 appref; /* remember, can't use globals */
    Int32 apptimeout;
    NetSocketRef udpsock;
    Int16 bytes, rv;
    unsigned char buf[32];
    UInt16 tp;
    UInt32 mxtime, curtime;
    NetFDSetType rfds, wfds;
    EventType e;
    Boolean ismail=false, ispacket=false;
    struct sPrefsR *prefsr;
    Int16 prefslocked=0;
    DmOpenRef maindb;
    UInt32 smstrig;

    if (almProcCmd != almProcCmdTriggered) return;

    do {

    e.eType = nilEvent; /* to make failure put it to sleep */

    /* TAKE ME OUT */
    ismail = true; break;

    /* we may be here because of sysAppLaunchCmdCustomBase+2 */
    if (!FtrGet(MainAppID, FTR_SMSTRIG, &smstrig)) {
        /* so we skip the udp probe and just do a pop */
        ismail = true;
        break;
    }

    tp = SysTicksPerSecond();
    apptimeout = 60*tp;

    /*
     * We need the POP host from our configuration database.  Make
     * sure to unlock this when I'm done.
     */
    if (!(prefslocked = lock_prefs_without_globals(&maindb, &prefsr))) break;

    /*
     * I send out a UDP probe to palmmail_udp on the mail host to
     * see if there is a need to POP.
     */
    err = SysLibFind("Net.lib", &appref);
    if (err) break;
    /*
     * This alarm is used to click-through any stupid dialogs that
     * stay on the screen waiting for "OK".
     */
    hit_ok_dialog();
    err = NetLibOpen(appref, &ifErrs);
    if ((err && err != netErrAlreadyOpen) || ifErrs) {
        /*
         * We always want to use a true on the NetLibClose() if there
         * was a failure with NetLibOpen().  Otherwise the PalmOS goes
         * berzerk thinking that it has a network connection when it
         * doesn't.. and it makes you cycle power to fix.
         */
        NetLibClose(appref, true);
        break;
    }
    if ((udpsock = connectsockudp(prefsr->POPPrefs.server, "14239", appref,
        apptimeout)) == 0) {
        NetLibClose(appref, false);
        break;
    }
    /* okay we have a good udp socket now */
    curtime = TimGetSeconds();
    MemSet((char *)buf, sizeof(buf), 0);
    buf[0] = 0xa5;
    buf[1] = 0x5a;
    MemMove(&buf[2], &curtime, sizeof(curtime));
    StrNCopy(&buf[7], prefsr->POPPrefs.user, sizeof(buf)-1);
    NetLibSend(appref, udpsock, buf, 16, 0, 0, 0, apptimeout, &err);
    /*
     * This waits for up to 10 seconds.  If the packet hasn't arrived yet,
     * or if the user does something, then drop out.
     */
    mxtime = TimGetSeconds()+10;
    do {
        netFDZero(&rfds);
        netFDZero(&wfds);
        netFDSet(sysFileDescStdIn, &rfds);
        netFDSet(udpsock, &rfds);
        rv = NetLibSelect(appref, netFDSetSize, &rfds, &wfds, NULL, tp,
            &err);
        if ((rv > 0) && netFDIsSet(udpsock, &rfds) &&
            ((bytes = NetLibReceive(appref, udpsock, buf,
            sizeof(buf), 0, 0, 0, 1, &err)) == 16) &&
            (buf[0] == 0xa5) && (buf[1] == 0x5a)) {
            ispacket = true;
            /* success */
            if (buf[6] == 1) ismail = true; /* now we'll do a full POP */
            break;
        }
        EvtGetEvent(&e, 0);
        SysHandleEvent(&e);
        curtime = TimGetSeconds();
    /* abort if timed out, or if user does something */
    } while ((curtime < mxtime) && (e.eType != keyDownEvent));
    /* check if user interrupted - and belay the mail POP */
    if ((!ispacket) && (curtime < mxtime)) ismail = false;
    NetLibSocketClose(appref, udpsock, apptimeout, &err);
    NetLibClose(appref, false); /* soft close if mail exists */
    } while (0); /* to catch all breaks */

    /* set this to allow for optimizing interrupted intervals */
    FtrSet(MainAppID, FTR_LASTTIME, paramP->alarmSeconds);

    /* unlock our prefs record if it was locked somewhere above */
    if (prefslocked == 1) unlock_prefs_without_globals(maindb, prefsr);

    if (!ismail) {
        EvtSetAutoOffTimer(SetExactly, 0); /* power off now */
    } else {
        /* set this feature so that I can prevent alarm stacking */
        FtrSet(MainAppID, FTR_POPACTIVE, 0);
        /*
         * Trigger the POP.  I would have done it all in here
         * except that PalmOS does something weird after this
         * Alarm Proc exits.  It restores the UI bits and messes
         * up things for me.
         */
        SetTimeOfNextAlarm(TimGetSeconds(), 0, 0);
    }
}

void DisplayAlarm(UInt16 launchFlags)
{
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType searchInfo;
    UInt32 res;
    UInt16 lf;
    UInt32 retval=999;
    UInt32 smstrig;
    FormActiveStateType formState;

    /* save all of the current app's UI state */
    FrmSaveActiveState(&formState);

    /* make sure I have time to do something */
    EvtSetAutoOffTimer(ResetTimer, 0);

    /*
     * Try to launch myself as a subroutine.
     */
    if ((DmGetNextDatabaseByTypeCreator(true, &searchInfo,
        sysFileTApplication, MainAppID, true, &cardNo, &dbID) == 0)) {
        lf = sysAppLaunchFlagNewStack;
        if (!(launchFlags&sysAppLaunchFlagSubCall)) {
            /* we're not in the app, so tell it to make globals. */
            lf |= sysAppLaunchFlagNewGlobals;
        }
        retval = SysAppLaunch(cardNo, dbID, lf,
            sysAppLaunchCmdCustomBase+1, NULL, &res);
    }

    /*
     * Tell the Mail application to refresh its list.  I'm doing this
     * here rather than in my applaunch because I want the UI to be
     * valid in case something weird happens.
     */
    mailapp_refresh_list();
    /*
     * I was going to reschedule the alarm if err, but I'm not doing
     * that now because I think it does more harm than good.
     * For example, I believe it will cause the Treo to bring
     * the radio out of sleep mode and begin a network search..
     * thereby causing excessive drain when we are clearly out
     * of coverage.
     *
     * 7/9/03 Well there is one annoying 0x1102 error that is driving
     * me nuts on my Treo 300, so I am going to reschedule on that
     * specific error.
     */
    if (res == 0x1102 /* Network busy */) {
        FtrSet(MainAppID, FTR_SMSTRIG, 120 /* seconds */);
        /* just don't unregister_sleep and it will occur again */
    } else {
        /* we may be here because of sysAppLaunchCmdCustomBase+2 */
        if (!FtrGet(MainAppID, FTR_SMSTRIG, &smstrig)) {
            /* no more alarms from this SMS */
            FtrUnregister(MainAppID, FTR_SMSTRIG);
            unregister_sleep();
            unregister_wake();
        }
    }

    /* get rid of our "alarm stacking" boolean */
    FtrUnregister(MainAppID, FTR_POPACTIVE);

    /* restore all of the current app's UI state */
    FrmRestoreActiveState(&formState);

    EvtSetAutoOffTimer(SetExactly, 5); /* power off in 5 seconds */
}

void AlarmTriggered(SysAlarmTriggeredParamType * cmdPBP)
{
    /* set this feature so that I can prevent alarm stacking */
    FtrSet(MainAppID, FTR_POPACTIVE, 0);
}

void hit_ok_dialog(void)
{
    FormPtr frm;
    ControlPtr ptr;
    UInt16 numobj;
    Int16 i1;
    char *lab;

    /*
     * I use this to click-through those annoying network
     * dialogs.
     */
    frm = FrmGetActiveForm();
    lab = (char *)FrmGetTitle(frm);
    if (lab &&
        (StrCaselessCompare("Service Connection Progress", lab) == 0)) {
        numobj = FrmGetNumberOfObjects(frm);
        for (i1=0; i1 < numobj; i1++) {
            ptr = FrmGetObjectPtr(frm, i1);
            if ((ptr->attr.enabled) && (ptr->style == buttonCtl)) {
                lab = (char *)CtlGetLabel(ptr);
                if (lab && (StrCaselessCompare("OK", lab) == 0)) {
                    CtlHitControl(ptr);
                }
            }
        }
    }
}
